import os
import re
import traceback
from datetime import datetime
from operator import attrgetter
from pathlib import Path
from threading import Thread
from typing import List, Set, Type, Tuple, Optional, Generator, Dict

from bauh.api import user
from bauh.api.abstract.controller import SearchResult, SoftwareManager, ApplicationContext, UpgradeRequirements, \
    UpgradeRequirement, TransactionResult, SoftwareAction, SettingsView, SettingsController
from bauh.api.abstract.disk import DiskCacheLoader
from bauh.api.abstract.handler import ProcessWatcher, TaskManager
from bauh.api.abstract.model import PackageHistory, PackageUpdate, SoftwarePackage, PackageSuggestion, \
    PackageStatus, CustomSoftwareAction, SuggestionPriority
from bauh.api.abstract.view import MessageType, FormComponent, SingleSelectComponent, InputOption, SelectViewType, \
    PanelComponent, ViewComponentAlignment
from bauh.commons import suggestions
from bauh.commons.boot import CreateConfigFile
from bauh.commons.html import strip_html, bold
from bauh.commons.system import ProcessHandler
from bauh.gems.flatpak import flatpak, CONFIG_FILE, UPDATES_IGNORED_FILE, FLATPAK_CONFIG_DIR, \
    EXPORTS_PATH, \
    get_icon_path, VERSION_1_5, VERSION_1_2, VERSION_1_12
from bauh.gems.flatpak.config import FlatpakConfigManager
from bauh.gems.flatpak.constants import FLATHUB_API_URL
from bauh.gems.flatpak.model import FlatpakApplication
from bauh.gems.flatpak.worker import FlatpakAsyncDataLoader

DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.000Z'
RE_INSTALL_REFS = re.compile(r'\d+\)\s+(.+)')


class FlatpakManager(SoftwareManager, SettingsController):

    def __init__(self, context: ApplicationContext):
        super(FlatpakManager, self).__init__(context=context)
        self.i18n = context.i18n
        self.api_cache = context.cache_factory.new(None)
        self.category_cache = context.cache_factory.new(None)
        context.disk_loader_factory.map(FlatpakApplication, self.api_cache)
        self.enabled = True
        self.http_client = context.http_client
        self.suggestions_cache = context.cache_factory.new(None)
        self.logger = context.logger
        self.configman = FlatpakConfigManager()
        self._action_full_update: Optional[CustomSoftwareAction] = None
        self._suggestions_file_url: Optional[str] = None

    def get_managed_types(self) -> Set["type"]:
        return {FlatpakApplication}

    def _map_to_model(self, app_json: dict, installed: bool, disk_loader: Optional[DiskCacheLoader], internet: bool = True) -> Tuple[FlatpakApplication, Optional[FlatpakAsyncDataLoader]]:

        app = FlatpakApplication(**app_json, i18n=self.i18n)
        app.installed = installed
        api_data = self.api_cache.get(app_json['id'])

        if app.runtime and app.latest_version is None:
            app.latest_version = app.version

        expired_data = api_data and api_data.get('expires_at') and api_data['expires_at'] <= datetime.utcnow()

        data_loader: Optional[FlatpakAsyncDataLoader] = None

        if not api_data or expired_data:
            if not app.runtime:
                if disk_loader:
                    disk_loader.fill(app)  # preloading cached disk data

                if internet:
                    data_loader = FlatpakAsyncDataLoader(app=app, api_cache=self.api_cache, manager=self,
                                                         context=self.context, category_cache=self.category_cache)
                    data_loader.start()

        else:
            app.fill_cached_data(api_data)
            app.status = PackageStatus.READY

        return app, data_loader

    def _get_search_remote(self) -> str:
        remotes = flatpak.list_remotes()

        if remotes['system']:
            remote_level = 'system'
        elif remotes['user']:
            remote_level = 'user'
        else:
            remote_level = 'user'
            ProcessHandler().handle_simple(flatpak.set_default_remotes(remote_level))

        return remote_level

    def search(self, words: str, disk_loader: DiskCacheLoader, limit: int = -1, is_url: bool = False) -> SearchResult:
        if is_url:
            return SearchResult([], [], 0)

        remote_level = self._get_search_remote()

        res = SearchResult([], [], 0)
        apps_found = flatpak.search(flatpak.get_version(), words, remote_level)

        if apps_found:
            already_read = set()
            installed_apps = self.read_installed(disk_loader=disk_loader, internet_available=True).installed

            if installed_apps:
                for app_found in apps_found:
                    for installed_app in installed_apps:
                        if app_found['id'] == installed_app.id:
                            res.installed.append(installed_app)
                            already_read.add(app_found['id'])

            if len(apps_found) > len(already_read):
                for app_found in apps_found:
                    if app_found['id'] not in already_read:
                        res.new.append(self._map_to_model(app_found, False, disk_loader)[0])

        res.total = len(res.installed) + len(res.new)
        return res

    def _add_updates(self, version: Tuple[str, ...], output: list):
        output.append(flatpak.list_updates_as_str(version))

    def _fill_required_runtimes(self, installation: str, output: List[Tuple[str, str]]):
        runtimes = flatpak.list_required_runtime_updates(installation=installation)

        if runtimes:
            output.extend(runtimes)

    def _fill_required_runtime_updates(self, output: Dict[str, List[Tuple[str, str]]]):
        threads = []
        for installation in ('system', 'user'):
            runtimes = list()
            output[installation] = runtimes
            t = Thread(target=self._fill_required_runtimes, args=(installation, runtimes))
            t.start()
            threads.append(t)

        for t in threads:
            t.join()

    def read_installed(self, disk_loader: Optional[DiskCacheLoader], limit: int = -1, only_apps: bool = False, pkg_types: Set[Type[SoftwarePackage]] = None,
                       internet_available: bool = None, wait_async_data: bool = False) -> SearchResult:
        version = flatpak.get_version()

        updates, required_runtimes = list(), dict()

        thread_updates, thread_runtimes = None, None
        if internet_available:
            thread_updates = Thread(target=self._add_updates, args=(version, updates))
            thread_updates.start()

            if version >= VERSION_1_12:
                thread_runtimes = Thread(target=self._fill_required_runtime_updates, args=(required_runtimes,))
                thread_runtimes.start()

        installed = flatpak.list_installed(version)

        update_map = None
        if thread_updates:
            thread_updates.join()
            update_map = updates[0]

        models = {}
        data_loaders: Optional[List[FlatpakAsyncDataLoader]] = [] if wait_async_data else None

        if installed:
            for app_json in installed:
                model, loader = self._map_to_model(app_json=app_json, installed=True,
                                                   disk_loader=disk_loader, internet=internet_available)
                model.update = False
                models[model.get_update_id(version)] = model

                if loader and data_loaders is not None:
                    data_loaders.append(loader)

        if update_map:
            for update_id in update_map['full']:
                model_with_update = models.get(update_id)
                if model_with_update:
                    model_with_update.update = True
                else:
                    # it is a new component that must be installed
                    update_id_split = update_id.split('/')
                    new_app = FlatpakApplication(id=update_id_split[0],
                                                 branch=update_id_split[1],
                                                 installation=update_id_split[2],
                                                 name=update_id_split[0].split('.')[-1].strip(),
                                                 version=update_id_split[1],
                                                 arch='x86_64' if self.context.is_system_x86_64() else 'x86',
                                                 origin=update_id_split[3] if len(update_id_split) == 4 else None)
                    new_app.update_component = True  # mark as "update component"
                    new_app.installed = True  # faking the "installed" status to be displayed as an update
                    new_app.update = True
                    new_app.update_ref()
                    models[update_id] = new_app

            if version >= VERSION_1_2:
                for partial_update_id in update_map['partial']:
                    partial_data = partial_update_id.split('/')

                    for model in models.values():
                        if model.installation == partial_data[2] and model.branch == partial_data[1]:
                            if model.id == partial_data[0]:
                                model.update = True
                                break
                            elif model.id in partial_data[0]:
                                partial_model = model.gen_partial(partial_data[0])
                                partial_model.update = True
                                models[partial_update_id] = partial_model
                                break

        if thread_runtimes:
            thread_runtimes.join()

        if required_runtimes:
            for installation in ('system', 'user'):
                installation_runtimes = required_runtimes.get(installation)

                if installation_runtimes:
                    for ref, origin in installation_runtimes:
                        ref_split = ref.split('/')
                        models[f'{installation}.'] = FlatpakApplication(id=ref_split[1],
                                                                        ref=ref,
                                                                        origin=origin,
                                                                        name=ref_split[1],
                                                                        version=ref_split[-1],
                                                                        latest_version=ref_split[-1],
                                                                        runtime=True,
                                                                        installation=installation,
                                                                        installed=False,
                                                                        update_component=True,
                                                                        update=True)

        if models:
            ignored = self._read_ignored_updates()

            if ignored:
                for model in models.values():
                    if model.get_update_ignore_key() in ignored:
                        model.updates_ignored = True

        if data_loaders:
            for loader in data_loaders:
                loader.join()

        return SearchResult([*models.values()], None, len(models))

    def downgrade(self, pkg: FlatpakApplication, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        if not self._make_exports_dir(watcher):
            return False

        watcher.change_progress(10)
        watcher.change_substatus(self.i18n['flatpak.downgrade.commits'])

        history = self.get_history(pkg, full_commit_str=True)

        # downgrade is not possible if the app current commit in the first one:
        if history.pkg_status_idx == len(history.history) - 1:
            watcher.show_message(self.i18n['flatpak.downgrade.impossible.title'],
                                 self.i18n['flatpak.downgrade.impossible.body'].format(bold(pkg.name)),
                                 MessageType.ERROR)
            return False

        commit = history.history[history.pkg_status_idx + 1]['commit']
        watcher.change_substatus(self.i18n['flatpak.downgrade.reverting'])
        watcher.change_progress(50)
        success, _ = ProcessHandler(watcher).handle_simple(flatpak.downgrade(app_ref=pkg.ref,
                                                                             commit=commit,
                                                                             installation=pkg.installation,
                                                                             root_password=root_password,
                                                                             version=flatpak.get_version()))
        watcher.change_progress(100)
        return success

    def clean_cache_for(self, pkg: FlatpakApplication):
        super(FlatpakManager, self).clean_cache_for(pkg)
        self.api_cache.delete(pkg.id)

    def upgrade(self, requirements: UpgradeRequirements, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        flatpak_version = flatpak.get_version()

        if not self._make_exports_dir(watcher):
            return False

        for req in requirements.to_upgrade:
            watcher.change_status("{} {} ({})...".format(self.i18n['manage_window.status.upgrading'], req.pkg.name, req.pkg.version))
            related, deps = False, False
            ref = req.pkg.ref

            if req.pkg.partial and flatpak_version < VERSION_1_5:
                related, deps = True, True
                ref = req.pkg.base_ref

            try:
                if req.pkg.update_component:
                    self.logger.info(f"Installing {req.pkg}")
                    res, _ = ProcessHandler(watcher).handle_simple(flatpak.install(app_id=ref,
                                                                                   installation=req.pkg.installation,
                                                                                   origin=req.pkg.origin,
                                                                                   version=flatpak_version))

                else:
                    self.logger.info(f"Updating {req.pkg}")
                    res, _ = ProcessHandler(watcher).handle_simple(flatpak.update(app_ref=ref,
                                                                                  installation=req.pkg.installation,
                                                                                  related=related,
                                                                                  deps=deps,
                                                                                  version=flatpak_version))

                watcher.change_substatus('')
                if not res:
                    self.logger.warning("Could not upgrade '{}'".format(req.pkg.id))
                    return False
            except Exception:
                watcher.change_substatus('')
                self.logger.error("An error occurred while upgrading '{}'".format(req.pkg.id))
                traceback.print_exc()
                return False

        watcher.change_substatus('')
        return True

    def uninstall(self, pkg: FlatpakApplication, root_password: Optional[str], watcher: ProcessWatcher, disk_loader: DiskCacheLoader) -> TransactionResult:

        if not self._make_exports_dir(watcher):
            return TransactionResult.fail()

        flatpak_version = flatpak.get_version()
        uninstalled, _ = ProcessHandler(watcher).handle_simple(flatpak.uninstall(pkg.ref, pkg.installation,
                                                                                 flatpak_version))

        if uninstalled:
            if self.suggestions_cache:
                self.suggestions_cache.delete(pkg.id)

            self.revert_ignored_update(pkg)
            return TransactionResult(success=True, installed=None, removed=[pkg])

        return TransactionResult.fail()

    def get_info(self, app: FlatpakApplication) -> dict:
        if app.installed:
            if app.update_component:
                app_info = {'id': app.id,
                            'name': app.name,
                            'branch': app.branch,
                            'installation': app.installation,
                            'origin': app.origin,
                            'arch': app.arch,
                            'ref': app.ref,
                            'type': 'runtime' if app.runtime else self.i18n['unknown']}
            else:
                version = flatpak.get_version()
                id_ = app.base_id if app.partial and version < VERSION_1_5 else app.id
                app_info = flatpak.get_app_info_fields(id_, app.branch, app.installation)

                if app.partial and version < VERSION_1_5:
                    app_info['id'] = app.id
                    app_info['ref'] = app.ref

                app_info['name'] = app.name
                app_info['type'] = 'runtime' if app.runtime else 'app'
                app_info['description'] = strip_html(app.description) if app.description else ''

                if app.installation:
                    app_info['installation'] = app.installation

                if app_info.get('installed'):
                    app_info['installed'] = app_info['installed'].replace('?', ' ')

            return app_info
        else:
            res = self.http_client.get_json('{}/apps/{}'.format(FLATHUB_API_URL, app.id))

            if res:
                if res.get('categories'):
                    res['categories'] = [c.get('name') for c in res['categories']]

                for to_del in ('screenshots', 'iconMobileUrl', 'iconDesktopUrl'):
                    if res.get(to_del):
                        del res[to_del]

                for to_strip in ('description', 'currentReleaseDescription'):
                    if res.get(to_strip):
                        res[to_strip] = strip_html(res[to_strip])

                for to_date in ('currentReleaseDate', 'inStoreSinceDate'):
                    if res.get(to_date):
                        try:
                            res[to_date] = datetime.strptime(res[to_date], DATE_FORMAT)
                        except Exception:
                            self.context.logger.error('Could not convert date string {} as {}'.format(res[to_date], DATE_FORMAT))
                            pass

                return res
            else:
                return {}

    def get_history(self, pkg: FlatpakApplication, full_commit_str: bool = False) -> PackageHistory:
        pkg.commit = flatpak.get_commit(pkg.id, pkg.branch, pkg.installation)
        pkg_commit = pkg.commit if pkg.commit else None

        if pkg_commit and not full_commit_str:
            pkg_commit = pkg_commit[0:8]

        commits = flatpak.get_app_commits_data(pkg.ref, pkg.origin, pkg.installation, full_str=full_commit_str)

        status_idx = 0
        commit_found = False

        if pkg_commit is None and len(commits) > 1 and commits[0]['commit'] == '(null)':
            del commits[0]
            pkg_commit = commits[0]
            commit_found = True

        if not commit_found:
            for idx, data in enumerate(commits):
                if data['commit'] == pkg_commit:
                    status_idx = idx
                    commit_found = True
                    break

        if not commit_found and pkg_commit and commits[0]['commit'] == '(null)':
            commits[0]['commit'] = pkg_commit

        return PackageHistory(pkg=pkg, history=commits, pkg_status_idx=status_idx)

    def _make_exports_dir(self, watcher: ProcessWatcher) -> bool:
        if not os.path.exists(EXPORTS_PATH):
            self.logger.info("Creating dir '{}'".format(EXPORTS_PATH))
            watcher.print('Creating dir {}'.format(EXPORTS_PATH))
            try:
                Path(EXPORTS_PATH).mkdir(parents=True, exist_ok=True)
            except Exception:
                watcher.print('Error while creating the directory {}'.format(EXPORTS_PATH))
                return False

        return True

    def install(self, pkg: FlatpakApplication, root_password: Optional[str], disk_loader: DiskCacheLoader, watcher: ProcessWatcher) -> TransactionResult:
        if not self.context.root_user:
            flatpak_config = self.configman.get_config()
            install_level = flatpak_config['installation_level']
        else:
            install_level = 'system'

        if install_level is not None:
            self.logger.info("Default Flaptak installation level defined: {}".format(install_level))

            if install_level not in ('user', 'system'):
                watcher.show_message(title=self.i18n['error'].capitalize(),
                                     body=self.i18n['flatpak.install.bad_install_level.body'].format(field=bold('installation_level'),
                                                                                                     file=bold(CONFIG_FILE)),
                                     type_=MessageType.ERROR)
                return TransactionResult(success=False, installed=[], removed=[])

            pkg.installation = install_level
        else:
            user_level = watcher.request_confirmation(title=self.i18n['flatpak.install.install_level.title'],
                                                      body=self.i18n['flatpak.install.install_level.body'].format(bold(pkg.name)),
                                                      confirmation_label=self.i18n['no'].capitalize(),
                                                      deny_label=self.i18n['yes'].capitalize())
            pkg.installation = 'user' if user_level else 'system'

        remotes = flatpak.list_remotes()

        handler = ProcessHandler(watcher)

        if pkg.installation == 'user' and not remotes['user']:
            handler.handle_simple(flatpak.set_default_remotes('user'))
        elif pkg.installation == 'system' and not remotes['system']:
            if user.is_root():
                handler.handle_simple(flatpak.set_default_remotes('system'))
            else:
                valid, user_password = watcher.request_root_password()
                if not valid:
                    watcher.print('Operation aborted')
                    return TransactionResult(success=False, installed=[], removed=[])
                else:
                    if not handler.handle_simple(flatpak.set_default_remotes('system', user_password))[0]:
                        watcher.show_message(title=self.i18n['error'].capitalize(),
                                             body=self.i18n['flatpak.remotes.system_flathub.error'],
                                             type_=MessageType.ERROR)
                        watcher.print("Operation cancelled")
                        return TransactionResult(success=False, installed=[], removed=[])

        # retrieving all installed so it will be possible to know the additional installed runtimes after the operation succeeds
        flatpak_version = flatpak.get_version()
        installed = flatpak.list_installed(flatpak_version)
        installed_by_level = {'{}:{}:{}'.format(p['id'], p['name'], p['branch']) for p in installed if p['installation'] == pkg.installation} if installed else None

        if not self._make_exports_dir(handler.watcher):
            return TransactionResult(success=False, installed=[], removed=[])

        installed, output = handler.handle_simple(flatpak.install(str(pkg.id), pkg.origin, pkg.installation,
                                                                  flatpak_version))

        if not installed and 'error: No ref chosen to resolve matches' in output:
            ref_opts = RE_INSTALL_REFS.findall(output)

            if ref_opts and len(ref_opts) > 1:
                view_opts = [InputOption(label=o, value=o.strip()) for o in ref_opts if o]
                ref_select = SingleSelectComponent(type_=SelectViewType.RADIO, options=view_opts, default_option=view_opts[0], label='')
                if watcher.request_confirmation(title=self.i18n['flatpak.install.ref_choose.title'],
                                                body=self.i18n['flatpak.install.ref_choose.body'].format(bold(pkg.name)),
                                                components=[ref_select],
                                                confirmation_label=self.i18n['proceed'].capitalize(),
                                                deny_label=self.i18n['cancel'].capitalize()):
                    ref = ref_select.get_selected()
                    installed, output = handler.handle_simple(flatpak.install(ref, pkg.origin, pkg.installation,
                                                                              flatpak_version))
                    pkg.ref = ref
                    pkg.runtime = 'runtime' in ref
                else:
                    watcher.print('Aborted by the user')
                    return TransactionResult.fail()
            else:
                return TransactionResult.fail()

        if installed:
            try:
                fields = flatpak.get_fields(str(pkg.id), pkg.branch, ['Ref', 'Branch'])

                if fields:
                    pkg.ref = fields[0]
                    pkg.branch = fields[1]
            except Exception:
                traceback.print_exc()

        if installed:
            new_installed = [pkg]
            current_installed = flatpak.list_installed(flatpak_version)
            current_installed_by_level = [p for p in current_installed if p['installation'] == pkg.installation] if current_installed else None

            if current_installed_by_level and (not installed_by_level or len(current_installed_by_level) > len(installed_by_level) + 1):
                pkg_key = '{}:{}:{}'.format(pkg.id, pkg.name, pkg.branch)
                net_available = self.context.is_internet_available()
                for p in current_installed_by_level:
                    current_key = '{}:{}:{}'.format(p['id'], p['name'], p['branch'])
                    if current_key != pkg_key and (not installed_by_level or current_key not in installed_by_level):
                        new_installed.append(self._map_to_model(app_json=p, installed=True,
                                                                disk_loader=disk_loader, internet=net_available)[0])

            return TransactionResult(success=installed, installed=new_installed, removed=[])
        else:
            return TransactionResult.fail()

    def is_enabled(self):
        return self.enabled

    def set_enabled(self, enabled: bool):
        self.enabled = enabled

    def can_work(self) -> Tuple[bool, Optional[str]]:
        return (True, None) if flatpak.is_installed() else (False, self.i18n['missing_dep'].format(dep=bold('flatpak')))

    def requires_root(self, action: SoftwareAction, pkg: FlatpakApplication) -> bool:
        return action == SoftwareAction.DOWNGRADE and pkg.installation == 'system'

    def prepare(self, task_manager: TaskManager, root_password: Optional[str], internet_available: bool):
        CreateConfigFile(taskman=task_manager, configman=self.configman, i18n=self.i18n,
                         task_icon_path=get_icon_path(), logger=self.logger).start()

    def list_updates(self, internet_available: bool) -> List[PackageUpdate]:
        updates = []
        installed = self.read_installed(None, internet_available=internet_available, wait_async_data=True).installed

        if installed:
            for p in installed:
                if isinstance(p, FlatpakApplication) and p.update and not p.is_update_ignored():
                    updates.append(PackageUpdate(pkg_id=f'{p.id}:{p.branch}:{p.installation}',
                                                 pkg_type='Flatpak',
                                                 name=p.name,
                                                 version=p.latest_version))

        return updates

    def list_warnings(self, internet_available: bool) -> Optional[List[str]]:
        pass

    def _read_local_suggestions_file(self) -> Optional[str]:
        try:
            with open(self.suggestions_file_url) as f:
                suggestions_str = f.read()

            return suggestions_str
        except FileNotFoundError:
            self.logger.error(f"Local Flatpak suggestions file not found: {self.suggestions_file_url}")
        except OSError:
            self.logger.error(f"Could not read local Flatpak suggestions file: {self.suggestions_file_url}")
            traceback.print_exc()

    def _download_remote_suggestions_file(self) -> Optional[str]:
        self.logger.info(f"Downloading the Flatpak suggestions from {self.suggestions_file_url}")
        file = self.http_client.get(self.suggestions_file_url)

        if file:
            return file.text

    def _fill_suggestion(self, appid: str, priority: SuggestionPriority, flatpak_version: Tuple[str, ...], remote: str,
                         output: List[PackageSuggestion]):
        app_json = flatpak.search(flatpak_version, appid, remote, app_id=True)

        if app_json:
            model = PackageSuggestion(self._map_to_model(app_json[0], False, None)[0], priority)
            self.suggestions_cache.add(appid, model)
            output.append(model)
        else:
            self.logger.warning(f"Could not find Flatpak suggestions '{appid}'")

    def list_suggestions(self, limit: int, filter_installed: bool) -> Optional[List[PackageSuggestion]]:
        if limit == 0:
            return

        if self.is_local_suggestions_file_mapped():
            suggestions_str = self._read_local_suggestions_file()
        else:
            suggestions_str = self._download_remote_suggestions_file()

        if suggestions_str is None:
            return

        if not suggestions_str:
            self.logger.warning(f"No Flatpak suggestion found in {self.suggestions_file_url}")
            return

        ids_prios = suggestions.parse(suggestions_str, self.logger, 'Flatpak')

        if not ids_prios:
            self.logger.warning(f"No Flatpak suggestion could be parsed from {self.suggestions_file_url}")
            return

        suggestion_by_priority = suggestions.sort_by_priority(ids_prios)

        if filter_installed:
            installed = {i.id for i in self.read_installed(disk_loader=None).installed}

            if installed:
                suggestion_by_priority = tuple(id_ for id_ in suggestion_by_priority if id_ not in installed)

        if suggestion_by_priority and 0 < limit < len(suggestion_by_priority):
            suggestion_by_priority = suggestion_by_priority[0:limit]

        self.logger.info(f'Available Flatpak suggestions: {len(suggestion_by_priority)}')

        if not suggestion_by_priority:
            return

        flatpak_version = flatpak.get_version()
        remote = self._get_search_remote()

        self.logger.info("Mapping Flatpak suggestions")
        res, fill_suggestions = [], []
        cached_count = 0

        for appid in suggestion_by_priority:
            cached_instance = self.suggestions_cache.get(appid)

            if cached_instance:
                res.append(cached_instance)
                cached_count += 1
            else:
                fill = Thread(target=self._fill_suggestion, args=(appid, ids_prios[appid], flatpak_version,
                                                                  remote, res))
                fill.start()
                fill_suggestions.append(fill)

        for fill in fill_suggestions:
            fill.join()

        if cached_count > 0:
            self.logger.info(f"Returning {cached_count} cached Flatpak suggestions")

        return res

    def is_default_enabled(self) -> bool:
        return True

    def launch(self, pkg: FlatpakApplication):
        flatpak.run(str(pkg.id))

    def get_screenshots(self, pkg: FlatpakApplication) -> Generator[str, None, None]:
        screenshots_url = f'{FLATHUB_API_URL}/apps/{pkg.id}'

        try:
            res = self.http_client.get_json(screenshots_url)

            if res and res.get('screenshots'):
                for s in res['screenshots']:
                    if s.get('imgDesktopUrl'):
                        yield s['imgDesktopUrl']

        except Exception as e:
            if e.__class__.__name__ == 'JSONDecodeError':
                self.context.logger.error("Could not decode json from '{}'".format(screenshots_url))
            else:
                traceback.print_exc()

    def get_settings(self) -> Optional[Generator[SettingsView, None, None]]:
        if not self.context.root_user:
            fields = []

            flatpak_config = self.configman.get_config()

            install_opts = [InputOption(label=self.i18n['flatpak.config.install_level.system'].capitalize(),
                                        value='system',
                                        tooltip=self.i18n['flatpak.config.install_level.system.tip']),
                            InputOption(label=self.i18n['flatpak.config.install_level.user'].capitalize(),
                                        value='user',
                                        tooltip=self.i18n['flatpak.config.install_level.user.tip']),
                            InputOption(label=self.i18n['ask'].capitalize(),
                                        value=None,
                                        tooltip=self.i18n['flatpak.config.install_level.ask.tip'].format(app=self.context.app_name))]
            fields.append(SingleSelectComponent(label=self.i18n['flatpak.config.install_level'],
                                                options=install_opts,
                                                default_option=[o for o in install_opts if o.value == flatpak_config['installation_level']][0],
                                                max_per_line=len(install_opts),
                                                type_=SelectViewType.COMBO,
                                                alignment=ViewComponentAlignment.CENTER,
                                                id_='install'))

            yield SettingsView(self, PanelComponent([FormComponent(fields, self.i18n['installation'].capitalize())]))

    def save_settings(self, component: PanelComponent) -> Tuple[bool, Optional[List[str]]]:
        flatpak_config = self.configman.get_config()
        form = component.get_component_by_idx(0, FormComponent)
        flatpak_config['installation_level'] = form.get_component('install', SingleSelectComponent).get_selected()

        try:
            self.configman.save_config(flatpak_config)
            return True, None
        except Exception:
            return False, [traceback.format_exc()]

    def get_upgrade_requirements(self, pkgs: List[FlatpakApplication], root_password: Optional[str], watcher: ProcessWatcher) -> UpgradeRequirements:
        flatpak_version = flatpak.get_version()

        user_pkgs, system_pkgs = [], []

        for pkg in pkgs:
            if pkg.installation == 'user':
                user_pkgs.append(pkg)
            else:
                system_pkgs.append(pkg)

        for apps_by_install in ((user_pkgs, 'user'), (system_pkgs, 'system')):
            if apps_by_install[0]:
                sizes = flatpak.map_update_download_size([str(p.id) for p in apps_by_install[0]], apps_by_install[1], flatpak_version)

                if sizes:
                    for p in apps_by_install[0]:
                        p.size = sizes.get(str(p.id))

        to_update = [UpgradeRequirement(pkg=p, extra_size=0, required_size=p.size) for p in self.sort_update_order(pkgs)]
        return UpgradeRequirements(None, None, to_update, [])

    def sort_update_order(self, pkgs: List[FlatpakApplication]) -> List[FlatpakApplication]:
        runtimes, apps = set(), set()

        for p in pkgs:
            if p.runtime:
                runtimes.add(p)
            else:
                apps.add(p)

        sorted_list = []
        for comps in (runtimes, apps):
            if comps:
                comp_list = list(comps)
                comp_list.sort(key=attrgetter('installation', 'name', 'id'))
                sorted_list.extend(comp_list)

        return sorted_list

    def _read_ignored_updates(self) -> Set[str]:
        ignored = set()
        if os.path.exists(UPDATES_IGNORED_FILE):
            with open(UPDATES_IGNORED_FILE) as f:
                ignored_txt = f.read()

            for l in ignored_txt.split('\n'):
                if l:
                    line_clean = l.strip()

                    if line_clean:
                        ignored.add(line_clean)

        return ignored

    def _write_ignored_updates(self, keys: Set[str]):
        Path(FLATPAK_CONFIG_DIR).mkdir(parents=True, exist_ok=True)
        ignored_list = [*keys]
        ignored_list.sort()

        with open(UPDATES_IGNORED_FILE, 'w+') as f:
            if ignored_list:
                for ignored in ignored_list:
                    f.write('{}\n'.format(ignored))
            else:
                f.write('')

    def ignore_update(self, pkg: FlatpakApplication):
        ignored_keys = self._read_ignored_updates()

        pkg_key = pkg.get_update_ignore_key()

        if pkg_key not in ignored_keys:
            ignored_keys.add(pkg_key)
            self._write_ignored_updates(ignored_keys)

        pkg.updates_ignored = True

    def revert_ignored_update(self, pkg: FlatpakApplication):
        ignored_keys = self._read_ignored_updates()

        if ignored_keys:
            pkg_key = pkg.get_update_ignore_key()

            if pkg_key in ignored_keys:
                ignored_keys.remove(pkg_key)
                self._write_ignored_updates(ignored_keys)

        pkg.updates_ignored = False

    def gen_custom_actions(self) -> Generator[CustomSoftwareAction, None, None]:
        yield self.action_full_update

    def full_update(self, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        handler = ProcessHandler(watcher)
        return handler.handle_simple(flatpak.full_update(flatpak.get_version()))[0]

    @property
    def action_full_update(self) -> CustomSoftwareAction:
        if self._action_full_update is None:
            self._action_full_update = CustomSoftwareAction(i18n_label_key='flatpak.action.full_update',
                                                            i18n_description_key='flatpak.action.full_update.description',
                                                            i18n_status_key='flatpak.action.full_update.status',
                                                            backup=True,
                                                            manager=self,
                                                            requires_internet=True,
                                                            icon_path=get_icon_path(),
                                                            manager_method='full_update',
                                                            requires_root=False)

        return self._action_full_update

    @property
    def suggestions_file_url(self) -> str:
        if self._suggestions_file_url is None:
            file_url = self.context.get_suggestion_url(self.__module__)

            if not file_url:
                file_url = 'https://raw.githubusercontent.com/vinifmor/bauh-files/master/flatpak/suggestions.txt'

            self._suggestions_file_url = file_url

            if file_url.startswith('/'):
                self.logger.info(f"Local Flatpak suggestions file mapped: {file_url}")

        return self._suggestions_file_url

    def is_local_suggestions_file_mapped(self) -> bool:
        return self.suggestions_file_url.startswith('/')



import json
import os
import re
from pathlib import Path
from typing import List, Dict, Tuple, Optional, Callable

from bauh.gems.arch import pacman
from bauh.gems.arch.model import ArchPackage

RE_DESKTOP_ENTRY = re.compile(r'[\n^](Exec|Icon|NoDisplay)\s*=\s*(.+)')
RE_CLEAN_NAME = re.compile(r'[+*?%]')


def write_several(pkgs: Dict[str, ArchPackage], overwrite: bool = True, maintainer: str = None, after_desktop_files: Optional[Callable] = None, after_written: Optional[Callable[[str], None]] = None) -> int:
    if overwrite:
        to_cache = {p.name for p in pkgs.values()}
    else:
        to_cache = {p.name for p in pkgs.values() if not os.path.exists(p.get_disk_cache_path())}

    desktop_files = pacman.map_desktop_files(*to_cache)

    if after_desktop_files:
        after_desktop_files()

    if not desktop_files:
        for pkgname in to_cache:
            write(pkg=pkgs[pkgname], maintainer=maintainer, after_written=after_written)

        return len(to_cache)
    else:
        for pkgname in to_cache:
            pkgfiles = desktop_files.get(pkgname)

            if not pkgfiles:
                write(pkg=pkgs[pkgname], maintainer=maintainer, after_written=after_written)
            else:
                desktop_entry = find_best_desktop_entry(pkgname, pkgfiles)

                if desktop_entry:
                    write(pkg=pkgs[pkgname], maintainer=maintainer, after_written=after_written,
                          desktop_file=desktop_entry[0], command=desktop_entry[1], icon=desktop_entry[2])
                else:
                    write(pkg=pkgs[pkgname], maintainer=maintainer, after_written=after_written)

        return len(to_cache)


def find_best_desktop_entry(pkgname: str, desktop_files: List[str]) -> Optional[Tuple[str, str, str]]:
    if len(desktop_files) == 1:
        exec_icon = read_desktop_exec_and_icon(pkgname, desktop_files[0])

        if exec_icon:
            return desktop_files[0], exec_icon[0], exec_icon[1]
    else:
        # trying to find the exact name match:
        for dfile in desktop_files:
            if dfile.endswith('{}.desktop'.format(pkgname)):
                exec_icon = read_desktop_exec_and_icon(pkgname, dfile)

                if exec_icon:
                    return dfile, exec_icon[0], exec_icon[1]

        # trying to find a close name match:
        clean_name = RE_CLEAN_NAME.sub('', pkgname)
        for dfile in desktop_files:
            if dfile.endswith('{}.desktop'.format(clean_name)):
                exec_icon = read_desktop_exec_and_icon(clean_name, dfile)

                if exec_icon:
                    return dfile, exec_icon[0], exec_icon[1]

        # finding any match:
        for dfile in desktop_files:
            exec_icon = read_desktop_exec_and_icon(pkgname, dfile)

            if exec_icon:
                return dfile, exec_icon[0], exec_icon[1]


def read_desktop_exec_and_icon(pkgname: str, desktop_file: str) -> Optional[Tuple[str, str]]:
    if os.path.isfile(desktop_file):
        with open(desktop_file) as f:
            possibilities = set()

            content = f.read()
            cmd, icon = None, None
            for field in RE_DESKTOP_ENTRY.findall(content):
                if field[0] == 'Exec':
                    cmd = field[1].strip().replace('"', '')
                elif field[0] == 'Icon':
                    icon = field[1].strip()
                elif field[0] == 'NoDisplay' and field[1].strip().lower() == 'true':
                    cmd, icon = None, None

                if cmd and icon:
                    possibilities.add((cmd, icon))
                    cmd, icon = None, None

            if possibilities:
                if len(possibilities) == 1:
                    return [*possibilities][0]
                else:
                    # trying to find the exact name x command match
                    for p in possibilities:
                        if p[0].startswith('{} '.format(pkgname)):
                            return p

                    return sorted(possibilities)[0]  # returning any possibility


def write(pkg: ArchPackage, desktop_file: Optional[str] = None, command: Optional[str] = None,
          icon: Optional[str] = None, maintainer: Optional[str] = None, after_written: Optional[callable] = None):
    pkg.desktop_entry = desktop_file
    pkg.command = command
    pkg.icon_path = icon

    if maintainer and not pkg.maintainer:
        pkg.maintainer = maintainer

    Path(pkg.get_disk_cache_path()).mkdir(parents=True, exist_ok=True)

    data = pkg.get_data_to_cache()

    with open(pkg.get_disk_data_path(), 'w+') as f:
        f.write(json.dumps(data))

    if after_written:
        after_written(pkg.name)



from typing import Optional

from bauh.commons.config import YAMLConfigManager
from bauh.gems.arch import CONFIG_FILE, get_pkgbuild_dir


def get_build_dir(arch_config: dict, user: Optional[str]) -> str:
    build_dir = arch_config.get('aur_build_dir')

    if not build_dir:
        build_dir = get_pkgbuild_dir(user)

    return build_dir


class ArchConfigManager(YAMLConfigManager):

    def __init__(self):
        super(ArchConfigManager, self).__init__(config_file_path=CONFIG_FILE)

    def get_default_config(self) -> dict:
        return {'optimize': True,
                "sync_databases": True,
                "clean_cached": True,
                'aur': True,
                'repositories': True,
                "refresh_mirrors_startup": False,
                "sync_databases_startup": True,
                'mirrors_sort_limit': 5,
                'repositories_mthread_download': False,
                'automatch_providers': True,
                'edit_aur_pkgbuild': False,
                'aur_build_dir': None,
                'aur_remove_build_dir': True,
                'aur_build_only_chosen': True,
                'check_dependency_breakage': True,
                'suggest_unneeded_uninstall': False,
                'suggest_optdep_uninstall': False,
                'aur_idx_exp': 1,
                'categories_exp': 24,
                'aur_rebuild_detector': False,
                "aur_rebuild_detector_no_bin": True,
                "prefer_repository_provider": True,
                'suggestions_exp': 24}



import logging
import time
import traceback
from threading import Thread
from typing import Dict, Set, List, Tuple, Iterable, Optional, Any

from bauh.api.abstract.controller import UpgradeRequirements, UpgradeRequirement
from bauh.api.abstract.handler import ProcessWatcher
from bauh.gems.arch import pacman, sorting
from bauh.gems.arch.aur import AURClient
from bauh.gems.arch.dependencies import DependenciesAnalyser
from bauh.gems.arch.exceptions import PackageNotFoundException
from bauh.gems.arch.model import ArchPackage
from bauh.gems.arch.pacman import RE_DEP_OPERATORS
from bauh.commons.version_util import match_required_version
from bauh.view.util.translation import I18n


class UpdateRequirementsContext:

    def __init__(self, to_update: Dict[str, ArchPackage], repo_to_update: Dict[str, ArchPackage],
                 aur_to_update: Dict[str, ArchPackage], repo_to_install: Dict[str, ArchPackage],
                 aur_to_install: Dict[str, ArchPackage], to_install: Dict[str, ArchPackage],
                 pkgs_data: Dict[str, dict], cannot_upgrade: Dict[str, UpgradeRequirement],
                 to_remove: Dict[str, UpgradeRequirement], installed: Dict[str, str],
                 provided_map: Dict[str, Set[str]], aur_index: Set[str], arch_config: dict,
                 remote_provided_map: Dict[str, Set[str]], remote_repo_map: Dict[str, str],
                 root_password: Optional[str], aur_supported: bool):
        self.to_update = to_update
        self.repo_to_update = repo_to_update
        self.aur_to_update = aur_to_update
        self.repo_to_install = repo_to_install
        self.aur_to_install = aur_to_install
        self.pkgs_data = pkgs_data
        self.cannot_upgrade = cannot_upgrade
        self.root_password = root_password
        self.installed = installed
        self.provided_map = provided_map
        self.to_remove = to_remove
        self.to_install = to_install
        self.aur_index = aur_index
        self.arch_config = arch_config
        self.remote_provided_map = remote_provided_map
        self.remote_repo_map = remote_repo_map
        self.aur_supported = aur_supported

    def update_provided_map(self, update: Dict[str, Set[str]]):
        if self.provided_map is None:
            self.provided_map = {**update}
        else:
            for provider, provided in update.items():
                provided_set = self.provided_map.get(provider)

                if provided_set is None:
                    provided_set = set()
                    self.provided_map[provider] = provided_set

                provided_set.update(provided)

    def add_to_provided_map(self, provider: str, provided: str):
        if self.provided_map is None:
            self.provided_map = dict()

        provided_set = self.provided_map.get(provider)

        if provided_set is None:
            provided_set = set()
            self.provided_map[provider] = provided_set

        provided_set.add(provided)


class UpdatesSummarizer:

    def __init__(self, aur_client: AURClient, i18n: I18n, logger: logging.Logger, deps_analyser: DependenciesAnalyser, aur_supported: bool, watcher: ProcessWatcher):
        self.aur_client = aur_client
        self.i18n = i18n
        self.logger = logger
        self.watcher = watcher
        self.deps_analyser = deps_analyser
        self.aur_supported = aur_supported

    def _fill_aur_pkg_update_data(self, pkg: ArchPackage, output: dict):
        output[pkg.name] = self.aur_client.map_update_data(pkg.get_base_name(), pkg.latest_version)

    def _handle_conflict_both_to_install(self, pkg1: str, pkg2: str, context: UpdateRequirementsContext):
        for src_pkg in {p for p, data in context.pkgs_data.items() if
                        data['d'] and pkg1 in data['d'] or pkg2 in data['d']}:
            if src_pkg not in context.cannot_upgrade:
                reason = self.i18n['arch.update_summary.to_install.dep_conflict'].format(f"'{pkg1}'", f"'{pkg2}'")
                context.cannot_upgrade[src_pkg] = UpgradeRequirement(context.to_update[src_pkg], reason)

            del context.to_update[src_pkg]

            if src_pkg in context.repo_to_update:
                del context.repo_to_update[src_pkg]
            else:
                del context.aur_to_update[src_pkg]

            del context.pkgs_data[src_pkg]

        for p in (pkg1, pkg2):
            if p in context.to_install:
                del context.to_install[p]

                if p in context.repo_to_install:
                    del context.repo_to_install[p]
                else:
                    del context.aur_to_install[p]

    def _handle_conflict_to_update_and_to_install(self, pkg1: str, pkg2: str, pkg1_to_install: bool, context: UpdateRequirementsContext):
        to_install, to_update = (pkg1, pkg2) if pkg1_to_install else (pkg2, pkg1)
        to_install_srcs = {p for p, data in context.pkgs_data.items() if data['d'] and to_install in data['d']}

        if to_update not in context.cannot_upgrade:
            srcs_str = ', '.join(("'{}'".format(p) for p in to_install_srcs))
            reason = self.i18n['arch.update_summary.to_update.conflicts_dep'].format("'{}'".format(to_install),
                                                                                     srcs_str)
            context.cannot_upgrade[to_install] = UpgradeRequirement(context.to_update[to_update], reason)

        if to_update in context.to_update:
            del context.to_update[to_update]

        for src_pkg in to_install_srcs:
            src_to_install = src_pkg in context.to_install
            pkg = context.to_install[src_pkg] if src_to_install else context.to_update[src_pkg]
            if src_pkg not in context.cannot_upgrade:
                reason = self.i18n['arch.update_summary.to_update.dep_conflicts'].format("'{}'".format(to_install),
                                                                                         "'{}'".format(to_update))
                context.cannot_upgrade[src_pkg] = UpgradeRequirement(pkg, reason)

            if src_to_install:
                del context.to_install[src_pkg]

                if src_pkg in context.repo_to_install:
                    del context.repo_to_install[src_pkg]
                else:
                    del context.aur_to_install[src_pkg]
            else:
                del context.to_update[src_pkg]

                if src_pkg in context.repo_to_update:
                    del context.repo_to_update[src_pkg]
                else:
                    del context.aur_to_update[src_pkg]

            del context.pkgs_data[src_pkg]

        if to_install in context.to_install:
            del context.to_install[to_install]

    def _handle_conflict_both_to_update(self, pkg1: str, pkg2: str, context: UpdateRequirementsContext):
        if pkg1 not in context.cannot_upgrade:
            reason = f"{self.i18n['arch.info.conflicts with'].capitalize()} '{pkg2}'"
            context.cannot_upgrade[pkg1] = UpgradeRequirement(pkg=context.to_update[pkg1], reason=reason)

        if pkg2 not in context.cannot_upgrade:
            reason = f"{self.i18n['arch.info.conflicts with'].capitalize()} '{pkg1}'"
            context.cannot_upgrade[pkg2] = UpgradeRequirement(pkg=context.to_update[pkg2], reason=reason)

        for p in (pkg1, pkg2):
            if p in context.to_update:
                del context.to_update[p]

                if p in context.repo_to_update:
                    del context.repo_to_update[p]
                else:
                    del context.aur_to_update[p]

    def _map_virtual_providers(self, providers: Dict[str, Set[str]], installed: Dict[str, str]) -> Dict[str, Set[str]]:
        ti = time.time()
        virtual_version = dict()
        for provider in providers:
            name_version = provider.split("=")
            if len(name_version) == 2 and name_version[0] not in installed:
                versions = virtual_version.get(name_version[0])

                if not versions:
                    versions = set()
                    virtual_version[name_version[0]] = versions

                versions.add(name_version[1])
        tf = time.time()
        self.logger.info(f"Took {tf - ti:.6f} seconds to map virtual providers of {len(providers)} packages")
        return virtual_version

    def _map_conflicts(self, data: Dict[str, Dict[str, Any]], providers: Dict[str, Set[str]],
                       versions: Dict[str, str]) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        Parameters
            pkgs_data: a dict mapping the packages whose conflicts need to be analyzed to their data
            providers: a dict mapping the available providers on the context (installed and to be installed) to their
            respective package
            versions: a dict mapping the package name to it's version (the updated or to be installed version)
        Return
            a tuple with two dictionaries:
                - first: containing all conflicts
                - second: containing mutual conflicts
        """
        virtual_providers = self._map_virtual_providers(providers, versions)

        root_conflict = {}
        mutual_conflicts = {}

        for pkg_name, data in data.items():
            if data['c']:
                for c in data['c']:
                    if c:
                        name_op_exp = DependenciesAnalyser.re_dep_operator().split(c)
                        conflict_name = name_op_exp[0]

                        if conflict_name != pkg_name:
                            conflict_providers = providers.get(conflict_name)
                            if conflict_providers:  # it means the conflict name matches a provided package
                                checked_conflicts = set()

                                if len(name_op_exp) == 1:  # if no expression is provided, add all providers
                                    checked_conflicts.update((p for p in conflict_providers if p != pkg_name))
                                else:
                                    virtual_versions = virtual_providers.get(conflict_name)

                                    if virtual_versions:
                                        # it means it's a virtual package
                                        # (e.g: 'xorg-server' provides a virtual package called 'x-server')
                                        for pversion in virtual_versions:
                                            if match_required_version(pversion, name_op_exp[1], name_op_exp[2]):
                                                # read the packages providing this specific virtual package version
                                                real_providers = providers.get(f"{conflict_name}={pversion}")

                                                if real_providers:
                                                    checked_conflicts.update(p for p in real_providers if p != pkg_name)

                                    else:
                                        for provider in conflict_providers:
                                            if provider != pkg_name:
                                                provider_version = versions.get(provider)
                                                if provider_version and match_required_version(provider_version,
                                                                                               name_op_exp[1],
                                                                                               name_op_exp[2]):
                                                    checked_conflicts.add(provider)

                                for provider in checked_conflicts:
                                    root_conflict[provider] = pkg_name

                                    if (pkg_name, provider) in root_conflict.items():
                                        mutual_conflicts[provider] = pkg_name

        return root_conflict, mutual_conflicts

    def _handle_mutual_conflicts(self, mutual_conflicts: Dict[str, str], all_conflicts: Dict[str, str],
                                 context: UpdateRequirementsContext):
        for pkg1, pkg2 in mutual_conflicts.items():
            pkg1_to_install = pkg1 in context.to_install
            pkg2_to_install = pkg2 in context.to_install

            if pkg1_to_install and pkg2_to_install:
                # remove both from to install and mark their source packages as 'cannot_update'
                self._handle_conflict_both_to_install(pkg1, pkg2, context)
            elif (pkg1_to_install and not pkg2_to_install) or (not pkg1_to_install and pkg2_to_install):
                self._handle_conflict_to_update_and_to_install(pkg1, pkg2, pkg1_to_install, context)
            else:
                # adding both to the 'cannot update' list
                self._handle_conflict_both_to_update(pkg1, pkg2, context)

        # removing conflicting packages from the packages selected to upgrade
        for pkg1, pkg2 in mutual_conflicts.items():
            for pkg_name in (pkg1, pkg2):
                if pkg_name in context.pkgs_data:
                    if context.pkgs_data[pkg_name].get('c'):
                        for c in context.pkgs_data[pkg_name]['c']:
                            # source = provided_map[c]
                            if c in all_conflicts:
                                del all_conflicts[c]

                    del context.pkgs_data[pkg_name]

    def _fill_conflicts(self, context: UpdateRequirementsContext, blacklist: Optional[Iterable[str]] = None):
        self.logger.info("Checking conflicts")

        conflicts, mutual_conflicts = self._map_conflicts(data=context.pkgs_data,
                                                          providers=context.provided_map,
                                                          versions=context.installed)

        if mutual_conflicts:
            self._handle_mutual_conflicts(mutual_conflicts, conflicts, context)

        if conflicts:
            for conflict_name, source_name in conflicts.items():
                if conflict_name not in context.to_remove and (not blacklist or conflict_name not in blacklist):
                    if conflict_name in context.to_update:
                        conflict = context.to_update[conflict_name]
                    else:
                        conflict = ArchPackage(name=conflict_name, installed=True, i18n=self.i18n)

                    reason = f"{self.i18n['arch.info.conflicts with'].capitalize()} '{source_name}'"
                    context.to_remove[conflict_name] = UpgradeRequirement(conflict, reason)

    def _map_and_add_package(self, pkg_data: Tuple[str, str], idx: int, output: dict):
        version = None

        if pkg_data[1] == 'aur':
            try:
                info = self.aur_client.get_src_info(pkg_data[0])

                if info:
                    version = info.get('pkgver')

                    if not version:
                        self.logger.warning("No version declared in SRCINFO of '{}'".format(pkg_data[0]))
                else:
                    self.logger.warning("Could not retrieve the SRCINFO for '{}'".format(pkg_data[0]))
            except:
                self.logger.warning("Could not retrieve the SRCINFO for '{}'".format(pkg_data[0]))
        else:
            version = pacman.get_version_for_not_installed(pkg_data[0])

        output[idx] = ArchPackage(name=pkg_data[0], version=version, latest_version=version, repository=pkg_data[1], i18n=self.i18n)

    def _fill_conflicts_to_install(self, context: UpdateRequirementsContext, install_data: Dict[str, Dict[str, Any]]):
        """
        Parameters
        context: update context
        install_data: a dict mapping the packages to be installed names by their data
        """
        # to properly fill conflicts considering new packages to be installed:
        # - the 'context.provided_map' should contain the providers of these new packages
        # - the 'context.installed' should contain the versions of these new packages
        provided_map_bkp = {**context.provided_map}
        self.__fill_provided_map(context=context, pkgs=context.to_install, fill_installed=False)

        # adding the new packages to install as 'installed'
        for pkg, data in install_data.items():
            context.installed[pkg] = data["v"]

        self._fill_conflicts(context, context.to_remove.keys())

        # restoring the original data structures
        context.provided_map = provided_map_bkp

        for pkg in install_data:
            if pkg in context.installed:
                del context.installed[pkg]

    def _fill_to_install(self, context: UpdateRequirementsContext) -> bool:
        ti = time.time()
        self.logger.info("Discovering updates missing packages")
        deps_data, deps_checked = {}, set()
        deps = self.deps_analyser.map_missing_deps(pkgs_data=context.pkgs_data,
                                                   provided_map=context.provided_map,
                                                   aur_index=context.aur_index,
                                                   deps_checked=deps_checked,
                                                   sort=True,
                                                   deps_data=deps_data,
                                                   remote_provided_map=context.remote_provided_map,
                                                   remote_repo_map=context.remote_repo_map,
                                                   watcher=self.watcher,
                                                   automatch_providers=context.arch_config['automatch_providers'],
                                                   prefer_repository_provider=context.arch_config['prefer_repository_provider'])

        if deps is None:
            tf = time.time()
            self.logger.info("It took {0:.2f} seconds to retrieve required upgrade packages".format(tf - ti))
            return False  # the user called the process off

        if deps:  # filtering selected packages
            selected_names = {p for p in context.to_update}
            deps = [dep for dep in deps if dep[0] not in selected_names]

            if deps:
                sorted_pkgs = {}
                aur_to_install_data = {}
                all_to_install_data = {}

                for idx, dep in enumerate(deps):
                    data = deps_data[dep[0]]
                    pkg = ArchPackage(name=dep[0], version=data['v'], latest_version=data['v'], repository=dep[1],
                                      i18n=self.i18n, package_base=data.get('b', dep[0]))
                    sorted_pkgs[idx] = pkg
                    context.to_install[dep[0]] = pkg

                    if pkg.repository == 'aur':
                        context.aur_to_install[pkg.name] = pkg
                        aur_to_install_data[pkg.name] = data
                    else:
                        context.repo_to_install[pkg.name] = pkg

                if context.repo_to_install:
                    all_to_install_data.update(pacman.map_updates_data(context.repo_to_install.keys()))

                if aur_to_install_data:
                    all_to_install_data.update(aur_to_install_data)

                if all_to_install_data:
                    context.pkgs_data.update(all_to_install_data)
                    self._fill_conflicts_to_install(context, all_to_install_data)

        if context.to_install:
            self.__fill_provided_map(context=context, pkgs=context.to_install, fill_installed=False)

        tf = time.time()
        self.logger.info(f"It took {tf - ti:.2f} seconds to retrieve required upgrade packages")
        return True

    def __fill_provided_map(self, context: UpdateRequirementsContext, pkgs: Dict[str, ArchPackage],
                            fill_installed: bool = True):
        if pkgs:
            ti = time.time()
            self.logger.info("Filling provided names")

            if not context.installed:
                context.installed = pacman.map_installed()

            installed_to_ignore = set()

            for pkgname in pkgs:
                context.add_to_provided_map(pkgname, pkgname)

                if fill_installed:
                    installed_to_ignore.add(pkgname)

                pdata = context.pkgs_data.get(pkgname)
                if pdata and pdata['p']:
                    context.add_to_provided_map(f"{pkgname}={pdata['v']}", pkgname)

                    ver_split = pdata['v'].split('-')

                    if len(ver_split) > 1:
                        context.add_to_provided_map(f"{pkgname}={'-'.join(ver_split[0:-1])}", pkgname)

                    for p in pdata['p']:
                        context.add_to_provided_map(p, pkgname)
                        split_provided = p.split('=')

                        if len(split_provided) > 1 and split_provided[0] != p:
                            context.add_to_provided_map(split_provided[0], pkgname)

            if context.installed and installed_to_ignore:  # filling the provided names of the installed
                installed_to_query = {*context.installed}.difference(installed_to_ignore)

                if installed_to_query:
                    context.update_provided_map(pacman.map_provided(remote=False, pkgs=installed_to_query))

            tf = time.time()
            self.logger.info("Filling provided names took {0:.2f} seconds".format(tf - ti))

    def __fill_aur_index(self, context: UpdateRequirementsContext):
        if context.aur_supported:
            self.logger.info("Loading AUR index")
            names = self.aur_client.read_index()

            if names:
                context.aur_index.update(names)
                self.logger.info("AUR index loaded on the context")

    def _map_requirement(self, pkg: ArchPackage, context: UpdateRequirementsContext,
                         installed_sizes: Optional[Dict[str, float]] = None, to_install: bool = False,
                         to_sync: Set[str] = None) -> UpgradeRequirement:

        requirement = UpgradeRequirement(pkg)

        if pkg.repository != 'aur':
            pkgdata = context.pkgs_data.get(pkg.name)

            if pkgdata:
                requirement.required_size = pkgdata['ds']
                requirement.extra_size = pkgdata['s']
    
                current_size = installed_sizes.get(pkg.name) if installed_sizes else None

                if current_size is not None and pkgdata['s'] is not None:
                    requirement.extra_size = pkgdata['s'] - current_size

            required_by = set()

            if to_install and to_sync and context.pkgs_data:
                names = pkgdata.get('p', {pkg.name}) if pkgdata else {pkg.name}
                to_sync_deps_cache = {}
                for p in to_sync:
                    if p != pkg.name and p in context.pkgs_data:
                        deps = to_sync_deps_cache.get(p)

                        if deps is None:
                            deps = context.pkgs_data[p]['d']

                            if deps is None:
                                deps = set()
                            else:
                                deps = {RE_DEP_OPERATORS.split(d)[0] for d in deps}

                            to_sync_deps_cache[p] = deps

                        if deps:
                            for n in names:
                                if n in deps:
                                    required_by.add(p)
                                    break

                requirement.reason = f"{self.i18n['arch.info.required by'].capitalize()}: " \
                                     f"{','.join(required_by) if required_by else '?'}"

        return requirement

    def summarize(self, pkgs: List[ArchPackage], root_password: Optional[str], arch_config: dict) \
            -> Optional[UpgradeRequirements]:

        res = UpgradeRequirements([], [], [], [])
        remote_provided_map = pacman.map_provided(remote=True)
        remote_repo_map = pacman.map_repositories()
        context = UpdateRequirementsContext(to_update={}, repo_to_update={}, aur_to_update={}, repo_to_install={},
                                            aur_to_install={}, to_install={}, pkgs_data={}, cannot_upgrade={},
                                            to_remove={}, installed=dict(), provided_map={}, aur_index=set(),
                                            arch_config=arch_config, root_password=root_password,
                                            remote_provided_map=remote_provided_map, remote_repo_map=remote_repo_map,
                                            aur_supported=self.aur_supported)
        self.__fill_aur_index(context)

        aur_data = {}
        aur_srcinfo_threads = []
        for p in pkgs:
            context.to_update[p.name] = p
            if p.repository == 'aur':
                context.aur_to_update[p.name] = p
                t = Thread(target=self._fill_aur_pkg_update_data, args=(p, aur_data), daemon=True)
                t.start()
                aur_srcinfo_threads.append(t)
            else:
                context.repo_to_update[p.name] = p

        if context.aur_to_update:
            for t in aur_srcinfo_threads:
                t.join()

        self.logger.info("Filling updates data")

        if context.repo_to_update:
            context.pkgs_data.update(pacman.map_updates_data(context.repo_to_update.keys()))

        if aur_data:
            context.pkgs_data.update(aur_data)

        self.__fill_provided_map(context=context, pkgs=context.to_update)

        if context.pkgs_data:
            self._fill_conflicts(context)

        try:
            if not self._fill_to_install(context):
                self.logger.info("The operation was cancelled by the user")
                return
        except PackageNotFoundException as e:
            self.logger.error(f"Package '{e.name}' not found")
            return

        if context.pkgs_data:
            self._fill_dependency_breakage(context)

        if context.to_remove:
            self.__update_context_based_on_to_remove(context)

        if context.to_update:
            installed_sizes = pacman.get_installed_size(list(context.to_update.keys()))

            sorted_pkgs = []

            if context.repo_to_update:
                # only sorting by name (pacman already knows the best order to perform the upgrade)
                sorted_pkgs.extend(context.repo_to_update.values())
                sorted_pkgs.sort(key=lambda pkg: pkg.name)

            if context.aur_to_update:  # adding AUR packages in the end
                sorted_aur = sorting.sort(context.aur_to_update.keys(), context.pkgs_data, context.provided_map)

                for aur_pkg in sorted_aur:
                    sorted_pkgs.append(context.aur_to_update[aur_pkg[0]])

            res.to_upgrade = [self._map_requirement(pkg, context, installed_sizes) for pkg in sorted_pkgs]

        if context.to_remove:
            res.to_remove = [p for p in context.to_remove.values()]

        if context.cannot_upgrade:
            res.cannot_upgrade = [d for d in context.cannot_upgrade.values()]

        if context.to_install:
            to_sync = {r.pkg.name for r in res.to_upgrade} if res.to_upgrade else {}
            to_sync.update(context.to_install.keys())
            res.to_install = [self._map_requirement(p, context, to_install=True, to_sync=to_sync)
                              for p in context.to_install.values()]

        res.context['data'] = context.pkgs_data
        return res

    def __update_context_based_on_to_remove(self, context: UpdateRequirementsContext):
        # filtering all package to synchronization from the transaction context
        to_sync = set()

        if context.to_update:
            to_sync.update(context.to_update.keys())

        if context.to_install:
            to_sync.update(context.to_install.keys())

        to_remove_provided = {}
        if to_sync:  # checking if any packages to sync on the context rely on the 'to remove' ones
            to_remove_provided.update(pacman.map_provided(remote=False, pkgs=context.to_remove.keys()))
            to_remove_from_sync = {}  # will store all packages that should be removed

            for pname in to_sync:
                if pname in context.pkgs_data:
                    deps = context.pkgs_data[pname].get('d')

                    if deps:
                        required = set()

                        for pkg in context.to_remove:
                            for provided in to_remove_provided[pkg]:
                                if provided in deps:
                                    required.add(pkg)
                                    break

                        if required:
                            to_remove_from_sync[pname] = required
                else:
                    self.logger.warning(f"Conflict resolution: package '{pname}' marked to synchronization "
                                        f"has no data loaded")

            if to_remove_from_sync:  # removing all these packages and their dependents from the context
                self._add_to_remove(to_sync, to_remove_from_sync, context)

        # checking if the installed packages that are not in the transaction context rely on the current
        # packages to be removed:
        current_to_remove = {*context.to_remove.keys()}
        required_by_to_remove = self.deps_analyser.map_all_required_by(current_to_remove, {*to_sync})

        if required_by_to_remove:
            # updating provided context:
            provided_not_mapped = set()
            for pkg in current_to_remove.difference({*to_remove_provided.keys()}):
                if pkg not in context.pkgs_data:
                    provided_not_mapped.add(pkg)
                else:
                    provided = context.pkgs_data[pkg].get('p')
                    if provided:
                        to_remove_provided[pkg] = provided
                    else:
                        provided_not_mapped.add(pkg)

            if provided_not_mapped:
                to_remove_provided.update(pacman.map_provided(remote=False, pkgs=provided_not_mapped))

            deps_no_data = {dep for dep in required_by_to_remove if dep not in context.pkgs_data}
            deps_nodata_deps = pacman.map_required_dependencies(*deps_no_data) if deps_no_data else {}

            reverse_to_remove_provided = {p: name for name, provided in to_remove_provided.items() for p in provided}

            for pkg in required_by_to_remove:
                if pkg not in context.to_remove:
                    if pkg in context.pkgs_data:
                        dep_deps = context.pkgs_data[pkg].get('d')
                    else:
                        dep_deps = deps_nodata_deps.get(pkg)

                    if dep_deps:
                        source = ', '.join(
                            (reverse_to_remove_provided[d] for d in dep_deps if d in reverse_to_remove_provided))
                        reason = f"{self.i18n['arch.info.depends on'].capitalize()} '{source if source else '?'}'"
                    else:
                        reason = '?'

                    pkg_repo = context.remote_repo_map.get(pkg)
                    pkg_version = context.installed.get(pkg)
                    context.to_remove[pkg] = UpgradeRequirement(pkg=ArchPackage(name=pkg,
                                                                                installed=True,
                                                                                i18n=self.i18n,
                                                                                version=pkg_version,
                                                                                latest_version=pkg_version,
                                                                                repository=pkg_repo),
                                                                reason=reason)

        for name in context.to_remove:  # upgrading lists
            if name in context.pkgs_data:
                del context.pkgs_data[name]

            if name in context.aur_to_update:
                del context.aur_to_update[name]

            if name in context.repo_to_update:
                del context.repo_to_update[name]

        removed_size = pacman.get_installed_size([*context.to_remove.keys()])

        if removed_size:
            for name, size in removed_size.items():
                if size is not None:
                    req = context.to_remove.get(name)
                    if req:
                        req.extra_size = size

    def _add_to_remove(self, pkgs_to_sync: Set[str], names: Dict[str, Set[str]], context: UpdateRequirementsContext,
                       to_ignore: Set[str] = None):

        blacklist = to_ignore if to_ignore else set()
        blacklist.update(names)

        dependents = {}
        for pname in pkgs_to_sync:
            if pname not in blacklist:
                data = context.pkgs_data.get(pname)

                if data:
                    deps = data.get('d')

                    if deps:
                        for n in names:
                            if n in deps:
                                all_deps = dependents.get(n, set())
                                all_deps.update(pname)
                                dependents[n] = all_deps

                else:
                    self.logger.warning(f"Package '{pname}' to sync could not be removed from the transaction context "
                                        f"because its data was not loaded")

        for n in names:
            if n in context.pkgs_data:
                if n not in context.to_remove:
                    depends_on = names.get(n)
                    if depends_on:
                        reason = f"{self.i18n['arch.info.depends on'].capitalize()} '{', '.join(depends_on)}'"
                    else:
                        reason = '?'

                    context.to_remove[n] = UpgradeRequirement(pkg=ArchPackage(name=n, installed=True, i18n=self.i18n),
                                                              reason=reason)

                all_deps = dependents.get(n)

                if all_deps:
                    self._add_to_remove(pkgs_to_sync, {dep: {n} for dep in all_deps}, context, blacklist)
            else:
                self.logger.warning(f"Package '{n}' could not be removed from the transaction context because its "
                                    f"data was not loaded")

    def _fill_dependency_breakage(self, context: UpdateRequirementsContext):
        if bool(context.arch_config['check_dependency_breakage']) and (context.to_update or context.to_install):
            ti = time.time()
            self.logger.info("Begin: checking dependency breakage")

            required_by = pacman.map_required_by(context.to_update.keys()) if context.to_update else {}

            if context.to_install:
                required_by.update(pacman.map_required_by(context.to_install.keys(), remote=True))

            reqs_not_in_transaction = set()
            reqs_in_transaction = set()

            transaction_pkgs = {*context.to_update.keys(), *context.to_install.keys()}

            for reqs in required_by.values():
                for r in reqs:
                    if r in transaction_pkgs:
                        reqs_in_transaction.add(r)
                    elif r in context.installed:
                        reqs_not_in_transaction.add(r)

            if not reqs_not_in_transaction and not reqs_in_transaction:
                return

            provided_versions = {}

            for p in context.provided_map:
                pkg_split = p.split('=')

                if len(pkg_split) > 1:
                    versions = provided_versions.get(pkg_split[0])

                    if versions is None:
                        versions = set()
                        provided_versions[pkg_split[0]] = versions

                    versions.add(pkg_split[1])

            if not provided_versions:
                return

            cannot_upgrade = set()

            for pkg, deps in pacman.map_required_dependencies(*reqs_not_in_transaction).items():
                self._add_dependency_breakage(pkgname=pkg,
                                              pkgdeps=deps,
                                              provided_versions=provided_versions,
                                              cannot_upgrade=cannot_upgrade,
                                              context=context)

            for pkg in reqs_in_transaction:
                data = context.pkgs_data[pkg]

                if data and data['d']:
                    self._add_dependency_breakage(pkgname=pkg,
                                                  pkgdeps=data['d'],
                                                  provided_versions=provided_versions,
                                                  cannot_upgrade=cannot_upgrade,
                                                  context=context)

            if cannot_upgrade:
                pkgs_available = {*context.to_update.values(), *context.to_install.values()}
                cannot_upgrade.update(self._add_dependents_as_cannot_upgrade(context=context,
                                                                             names=cannot_upgrade,
                                                                             pkgs_available=pkgs_available))

                for p in cannot_upgrade:
                    if p in context.to_update:
                        del context.to_update[p]

                    if p in context.repo_to_update:
                        del context.repo_to_update[p]

                    if p in context.aur_to_update:
                        del context.aur_to_update[p]

                    if p in context.pkgs_data:
                        del context.pkgs_data[p]

                    if p in context.to_install:
                        del context.to_install[p]

                    if p in context.repo_to_install:
                        del context.repo_to_install[p]

                    if p in context.aur_to_install:
                        del context.aur_to_install[p]

            tf = time.time()
            self.logger.info("End: checking dependency breakage. Time: {0:.2f} seconds".format(tf - ti))

    def _add_dependents_as_cannot_upgrade(self, context: UpdateRequirementsContext, names: Iterable[str],
                                          pkgs_available: Set[ArchPackage], already_removed: Optional[Set[str]] = None,
                                          iteration_level: int = 0) -> Set[str]:
        removed = set() if already_removed is None else already_removed
        removed.update(names)

        available = {p for p in pkgs_available if p.name not in removed}
        to_remove = set()

        if available:
            for pkg in available:
                if pkg.name not in removed:
                    data = context.pkgs_data.get(pkg.name)

                    if data and data['d']:
                        for dep in data['d']:
                            dep_providers = context.provided_map.get(dep)

                            if dep_providers:
                                for p in dep_providers:
                                    if p in names:
                                        to_remove.add(pkg.name)

                                        if pkg.name not in context.cannot_upgrade:
                                            reason = f"{self.i18n['arch.info.depends on'].capitalize()} {p}"
                                            req = UpgradeRequirement(pkg=pkg, reason=reason,
                                                                     sorting_priority=iteration_level - 1)
                                            context.cannot_upgrade[pkg.name] = req
                                        break

            if to_remove:
                removed.update(to_remove)
                self._add_dependents_as_cannot_upgrade(context=context, names=to_remove, pkgs_available=available,
                                                       already_removed=to_remove, iteration_level=iteration_level-1)

        return to_remove

    def _add_dependency_breakage(self, pkgname: str, pkgdeps: Optional[Set[str]],
                                 provided_versions: Dict[str, Set[str]], cannot_upgrade: Set[str],
                                 context: UpdateRequirementsContext):
        if pkgdeps:
            for dep in pkgdeps:
                dep_split = RE_DEP_OPERATORS.split(dep)

                if len(dep_split) > 1 and dep_split[1]:
                    real_providers = context.provided_map.get(dep_split[0])

                    if real_providers:
                        versions = provided_versions.get(dep_split[0])

                        if versions:
                            op = ''.join(RE_DEP_OPERATORS.findall(dep))

                            version_match = False

                            for v in versions:
                                try:
                                    if match_required_version(current_version=v,
                                                              operator=op,
                                                              required_version=dep_split[1]):
                                        version_match = True
                                        break
                                except:
                                    self.logger.error(f"Error when comparing versions {v} (provided) and "
                                                      f"{dep_split[1]} (required)")
                                    traceback.print_exc()

                            if not version_match:
                                for pname in real_providers:
                                    if pname not in cannot_upgrade:
                                        provider = context.to_update.get(pname)
                                        if provider:
                                            cannot_upgrade.add(pname)
                                            reason = self.i18n['arch.sync.dep_breakage.reason'].format(pkgname, dep)
                                            context.cannot_upgrade[pname] = UpgradeRequirement(pkg=provider,
                                                                                               reason=reason)



from typing import List, Set, Optional, Iterable, Tuple

from bauh.api.abstract.model import SoftwarePackage, CustomSoftwareAction
from bauh.commons import resource
from bauh.gems.arch import ROOT_DIR, ARCH_CACHE_DIR
from bauh.view.util.translation import I18n


class ArchPackage(SoftwarePackage):

    __action_allow_rebuild_check: Optional[CustomSoftwareAction] = None
    __action_enable_pkgbuild_edition: Optional[CustomSoftwareAction] = None
    __action_disable_pkgbuild_edition: Optional[CustomSoftwareAction] = None
    __action_reinstall: Optional[CustomSoftwareAction] = None
    __action_ignore_rebuild_check: Optional[CustomSoftwareAction] = None
    __cached_attrs: Optional[Tuple[str, ...]] = None
    __dynamic_categories: Optional[Tuple[str, ...]] = None

    @classmethod
    def action_allow_rebuild_check(cls) -> CustomSoftwareAction:
        if not cls.__action_allow_rebuild_check:
            cls.__action_allow_rebuild_check = CustomSoftwareAction(i18n_label_key='arch.action.rebuild_check.allow',
                                                                    i18n_status_key='arch.action.rebuild_check.allow.status',
                                                                    i18n_confirm_key='arch.action.rebuild_check.allow.confirm',
                                                                    i18n_description_key='arch.action.rebuild_check.allow.desc',
                                                                    requires_root=False,
                                                                    manager_method='set_rebuild_check',
                                                                    icon_path=resource.get_path('img/check.svg', ROOT_DIR))

        return cls.__action_allow_rebuild_check

    @classmethod
    def action_enable_pkgbuild_edition(cls) -> CustomSoftwareAction:
        if not cls.__action_enable_pkgbuild_edition:
            cls.__action_enable_pkgbuild_edition = CustomSoftwareAction(i18n_label_key='arch.action.enable_pkgbuild_edition',
                                                                        i18n_status_key='arch.action.enable_pkgbuild_edition.status',
                                                                        i18n_confirm_key='arch.action.enable_pkgbuild_edition.confirm',
                                                                        i18n_description_key='arch.action.enable_pkgbuild_edition.desc',
                                                                        requires_root=False,
                                                                        manager_method='enable_pkgbuild_edition',
                                                                        icon_path=resource.get_path('img/mark_pkgbuild.svg', ROOT_DIR))

        return cls.__action_enable_pkgbuild_edition

    @classmethod
    def action_disable_pkgbuild_edition(cls) -> CustomSoftwareAction:
        if not cls.__action_disable_pkgbuild_edition:
            cls.__action_disable_pkgbuild_edition = CustomSoftwareAction(i18n_label_key='arch.action.disable_pkgbuild_edition',
                                                                         i18n_status_key='arch.action.disable_pkgbuild_edition',
                                                                         i18n_confirm_key='arch.action.disable_pkgbuild_edition.confirm',
                                                                         i18n_description_key='arch.action.disable_pkgbuild_edition.desc',
                                                                         requires_root=False,
                                                                         manager_method='disable_pkgbuild_edition',
                                                                         icon_path=resource.get_path('img/unmark_pkgbuild.svg', ROOT_DIR))

        return cls.__action_disable_pkgbuild_edition

    @classmethod
    def action_reinstall(cls) -> CustomSoftwareAction:
        if not cls.__action_reinstall:
            cls.__action_reinstall = CustomSoftwareAction(i18n_label_key='arch.action.reinstall',
                                                          i18n_status_key='arch.action.reinstall.status',
                                                          i18n_confirm_key='arch.action.reinstall.confirm',
                                                          i18n_description_key='arch.action.reinstall.desc',
                                                          requires_root=True,
                                                          manager_method='reinstall',
                                                          icon_path=resource.get_path('img/build.svg', ROOT_DIR))

        return cls.__action_reinstall

    @classmethod
    def action_ignore_rebuild_check(cls) -> CustomSoftwareAction:
        if not cls.__action_ignore_rebuild_check:
            cls.__action_ignore_rebuild_check = CustomSoftwareAction(i18n_label_key='arch.action.rebuild_check.ignore',
                                                                     i18n_status_key='arch.action.rebuild_check.ignore.status',
                                                                     i18n_confirm_key='arch.action.rebuild_check.ignore.confirm',
                                                                     i18n_description_key='arch.action.rebuild_check.ignore.desc',
                                                                     requires_root=False,
                                                                     manager_method='set_rebuild_check',
                                                                     icon_path=resource.get_path('img/check_disabled.svg', ROOT_DIR))

        return cls.__action_ignore_rebuild_check

    @classmethod
    def cached_attrs(cls) -> Tuple[str, ...]:
        if cls.__cached_attrs is None:
            cls.__cached_attrs = ('command', 'icon_path', 'repository', 'maintainer', 'desktop_entry', 'categories',
                                  'last_modified', 'commit')

        return cls.__cached_attrs

    @classmethod
    def dynamic_categories(cls) -> Tuple[str, ...]:
        if cls.__dynamic_categories is None:
            cls.__dynamic_categories = ('orphan', 'out_of_date')

        return cls.__dynamic_categories

    def __init__(self, name: str = None, version: str = None, latest_version: str = None, description: str = None,
                 package_base: str = None, votes: int = None, popularity: float = None,
                 first_submitted: Optional[int] = None, last_modified: Optional[int] = None,
                 maintainer: str = None, url_download: str = None, pkgbuild: str = None, repository: str = None,
                 desktop_entry: str = None, installed: bool = False, srcinfo: dict = None, dependencies: Set[str] = None,
                 categories: List[str] = None, i18n: I18n = None, update_ignored: bool = False, arch: str = None,
                 pkgbuild_editable: bool = None, install_date: Optional[int] = None, commit: Optional[str] = None,
                 require_rebuild: bool = False, allow_rebuild: Optional[bool] = None, aur_update: bool = False,
                 out_of_date: Optional[bool] = None):

        super(ArchPackage, self).__init__(name=name, version=version, latest_version=latest_version, description=description,
                                          installed=installed, categories=categories)
        self.package_base = package_base
        self.votes = votes
        self.popularity = popularity
        self.maintainer = maintainer if maintainer else (repository if repository != 'aur' else None)
        self.url_download = url_download
        self.first_submitted = first_submitted
        self.last_modified = last_modified
        self.pkgbuild = pkgbuild
        self.repository = repository
        self.command = None
        self.icon_path = None
        self.desktop_entry = desktop_entry
        self.src_info = srcinfo
        self.dependencies = dependencies
        self.arch = arch
        self.i18n = i18n
        self.update_ignored = update_ignored
        self.view_name = name  # name displayed on the view
        self.pkgbuild_editable = pkgbuild_editable  # if the PKGBUILD can be edited by the user (only for AUR)
        self.install_date = install_date
        self.commit = commit  # only for AUR for downgrading purposes
        self.require_rebuild = require_rebuild
        self.allow_rebuild = allow_rebuild
        self.aur_update = aur_update
        self.out_of_date = out_of_date

    @staticmethod
    def disk_cache_path(pkgname: str):
        return f'{ARCH_CACHE_DIR}/installed/{pkgname}'

    def get_pkg_build_url(self):
        if self.package_base:
            return 'https://aur.archlinux.org/cgit/aur.git/plain/PKGBUILD?h=' + self.package_base

    def has_history(self):
        return self.can_be_downgraded()

    def has_info(self):
        return True

    def can_be_installed(self) -> bool:
        if super(ArchPackage, self).can_be_installed():
            return bool(self.url_download) if self.repository == 'aur' else True

    def can_be_downgraded(self):
        return self.installed and self.repository == 'aur'

    def get_type(self):
        return 'aur' if self.repository == 'aur' else 'arch_repo'

    def get_default_icon_path(self) -> str:
        return self.get_type_icon_path()

    def get_disk_icon_path(self) -> str:
        return self.icon_path

    def get_type_icon_path(self):
        return resource.get_path('img/{}.svg'.format('arch' if self.get_type() == 'aur' else 'repo'), ROOT_DIR)

    def is_application(self):
        return self.can_be_run()

    def get_base_name(self) -> str:
        return self.package_base if self.package_base else self.name

    def supports_disk_cache(self):
        return True

    def get_disk_cache_path(self) -> str:
        if self.name:
            return self.disk_cache_path(self.name)

    def get_data_to_cache(self) -> dict:
        cache = {}

        # required attrs to cache
        for a in self.cached_attrs():
            val = getattr(self, a)

            if val:
                if a == 'categories' and isinstance(val, list):
                    for cat in self.dynamic_categories():
                        if cat in val:
                            val.remove(cat)

                cache[a] = val

        return cache

    def fill_cached_data(self, data: dict):
        if data:
            for a in self.cached_attrs():
                val = data.get(a)
                if val:
                    setattr(self, a, val)

                    if a == 'icon_path':
                        self.icon_url = val

    def can_be_run(self) -> bool:
        # only returns if there is a desktop entry set for the application to avoid running command-line applications
        return bool(self.desktop_entry) and bool(self.command)

    def get_publisher(self):
        return self.maintainer

    def set_icon(self, paths: List[str]):
        self.icon_path = paths[0]

        if len(paths) > 1:
            for path in paths:
                if '/' in path:
                    self.icon_path = path
                    break

        self.icon_url = self.icon_path

    def has_screenshots(self) -> bool:
        return False

    def get_name_tooltip(self) -> str:
        return '{} ( {}: {} )'.format(self.name, self.i18n['repository'], self.repository)

    def supports_backup(self) -> bool:
        return True

    def is_update_ignored(self) -> bool:
        return self.update_ignored

    def supports_ignored_updates(self) -> bool:
        return self.installed

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '{} (name={}, command={}, icon_path={})'.format(self.__class__.__name__, self.name, self.command, self.icon_path)

    def __eq__(self, other):
        if isinstance(other, ArchPackage):
            if self.view_name is not None and other.view_name is not None:
                return self.view_name == other.view_name and self.repository == other.repository

            return self.name == other.name and self.repository == other.repository

    def get_cached_pkgbuild_path(self) -> str:
        return '{}/PKGBUILD'.format(self.get_disk_cache_path())

    def get_custom_actions(self) -> Optional[Iterable[CustomSoftwareAction]]:
        if self.installed and self.repository == 'aur':
            actions = [self.action_reinstall()]

            if self.pkgbuild_editable is not None:
                actions.append(self.action_disable_pkgbuild_edition() if self.pkgbuild_editable else self.action_enable_pkgbuild_edition())

            if self.allow_rebuild is not None:
                actions.append(self.action_ignore_rebuild_check() if self.allow_rebuild else self.action_allow_rebuild_check())

            return actions

    def get_update_tip(self) -> Optional[str]:
        if self.repository == 'aur' and self.allow_rebuild and self.require_rebuild:
            return self.i18n['arch.package.requires_rebuild'] + ' (rebuild)'

    def update_state(self):
        if self.repository == 'aur':
            if self.allow_rebuild and self.require_rebuild:
                self.update = True
            else:
                self.update = self.aur_update

    def __hash__(self):
        if self.view_name is not None:
            return hash((self.view_name, self.repository))
        else:
            return hash((self.name, self.repository))

    @property
    def orphan(self) -> bool:
        return self.maintainer is None



import multiprocessing
import os
import traceback
from pwd import getpwnam
from typing import Callable, Optional, TypeVar

R = TypeVar('R')


class CallAsUser:

    def __init__(self, target: Callable[[], R], user: str):
        self._target = target
        self._user = user

    def __call__(self, *args, **kwargs) -> R:
        try:
            os.setuid(getpwnam(self._user).pw_uid)
            return self._target()
        except Exception:
            traceback.print_exc()


class WriteToFile:

    def __init__(self, file_path: str, content: str):
        self._file_path = file_path
        self._content = content

    def __call__(self, *args, **kwargs) -> bool:
        try:
            with open(self._file_path, 'w+') as f:
                f.write(self._content)

            return True
        except Exception:
            traceback.print_exc()
            return False


def exec_as_user(target: Callable[[], R], user: Optional[str] = None) -> R:
    if user:
        with multiprocessing.Pool(1) as pool:
            return pool.apply(CallAsUser(target, user))
    else:
        return target()


def write_as_user(content: str, file_path: str, user: Optional[str] = None) -> bool:
    return exec_as_user(target=WriteToFile(file_path=file_path, content=content),
                        user=user)



import glob
import json
import logging
import os
import tarfile
import time
import traceback
from datetime import datetime, timedelta
from pathlib import Path
from threading import Thread
from typing import Optional, Generator

import requests

from bauh.api.abstract.handler import TaskManager, ProcessWatcher
from bauh.api.http import HttpClient
from bauh.commons.boot import CreateConfigFile
from bauh.commons.html import bold
from bauh.gems.appimage import get_icon_path, INSTALLATION_DIR, SYMLINKS_DIR, util, DATABASES_TS_FILE, \
    APPIMAGE_CACHE_DIR, DATABASE_APPS_FILE, DATABASE_RELEASES_FILE, URL_COMPRESSED_DATABASES
from bauh.gems.appimage.model import AppImage
from bauh.view.util.translation import I18n


class DatabaseUpdater(Thread):
    COMPRESS_FILE_PATH = f'{APPIMAGE_CACHE_DIR}/db.tar.gz'

    def __init__(self, i18n: I18n, http_client: HttpClient, logger: logging.Logger, taskman: TaskManager,
                 watcher: Optional[ProcessWatcher] = None, appimage_config: Optional[dict] = None, create_config: Optional[CreateConfigFile] = None):
        super(DatabaseUpdater, self).__init__(daemon=True)
        self.http_client = http_client
        self.logger = logger
        self.i18n = i18n
        self.taskman = taskman
        self.watcher = watcher
        self.task_id = 'appim_db'
        self.config = appimage_config
        self.create_config = create_config
        self.taskman.register_task(self.task_id, self.i18n['appimage.task.db_update'], get_icon_path())

    def should_update(self, appimage_config: dict) -> bool:
        ti = time.time()

        try:
            db_exp = int(appimage_config['database']['expiration'])
        except ValueError:
            self.logger.error("Could not parse settings property 'database.expiration': {}".format(appimage_config['database']['expiration']))
            return True

        if db_exp <= 0:
            self.logger.info("No expiration time configured for the AppImage database")
            return True

        files = {*glob.glob(f'{APPIMAGE_CACHE_DIR}/*')}

        if not files:
            self.logger.warning(f'No database files on {APPIMAGE_CACHE_DIR}')
            return True

        if DATABASES_TS_FILE not in files:
            self.logger.warning("No database timestamp file found ({})".format(DATABASES_TS_FILE))
            return True

        if DATABASE_APPS_FILE not in files:
            self.logger.warning("Database file '{}' not found".format(DATABASE_APPS_FILE))
            return True

        if DATABASE_RELEASES_FILE not in files:
            self.logger.warning("Database file '{}' not found".format(DATABASE_RELEASES_FILE))
            return True

        with open(DATABASES_TS_FILE) as f:
            dbs_ts_str = f.read()

        try:
            dbs_timestamp = datetime.fromtimestamp(float(dbs_ts_str))
        except Exception:
            self.logger.error('Could not parse the databases timestamp: {}'.format(dbs_ts_str))
            traceback.print_exc()
            return True

        update = dbs_timestamp + timedelta(minutes=db_exp) <= datetime.utcnow()
        self.logger.info('Finished. Took {0:.2f} seconds'.format(time.time() - ti))
        return update

    def _update_task_progress(self, progress: float, substatus: Optional[str] = None):
        self.taskman.update_progress(self.task_id, progress, substatus)

        if self.watcher:
            self.watcher.change_substatus(substatus)

    def download_databases(self) -> bool:
        self._update_task_progress(10, self.i18n['appimage.update_database.downloading'])
        self.logger.info('Retrieving AppImage databases')

        database_timestamp = datetime.utcnow().timestamp()
        try:
            res = self.http_client.get(URL_COMPRESSED_DATABASES, session=False)
        except Exception as e:
            self.logger.error("An error ocurred while downloading the AppImage database: {}".format(e.__class__.__name__))
            res = None

        if not res:
            self.logger.warning('Could not download the database file {}'.format(URL_COMPRESSED_DATABASES))
            return False

        Path(APPIMAGE_CACHE_DIR).mkdir(parents=True, exist_ok=True)

        with open(self.COMPRESS_FILE_PATH, 'wb+') as f:
            f.write(res.content)

        self.logger.info("Database file saved at {}".format(self.COMPRESS_FILE_PATH))

        self._update_task_progress(50, self.i18n['appimage.update_database.deleting_old'])
        old_db_files = glob.glob(f'{APPIMAGE_CACHE_DIR}/*.db')

        if old_db_files:
            self.logger.info('Deleting old database files')
            for f in old_db_files:
                os.remove(f)

            self.logger.info('Old database files deleted')

        self._update_task_progress(75, self.i18n['appimage.update_database.uncompressing'])
        self.logger.info('Uncompressing {}'.format(self.COMPRESS_FILE_PATH))

        try:
            tf = tarfile.open(self.COMPRESS_FILE_PATH)
            tf.extractall(APPIMAGE_CACHE_DIR)
            self.logger.info('Successfully uncompressed file {}'.format(self.COMPRESS_FILE_PATH))
        except Exception:
            self.logger.error('Could not extract file {}'.format(self.COMPRESS_FILE_PATH))
            traceback.print_exc()
            return False
        finally:
            self.logger.info('Deleting {}'.format(self.COMPRESS_FILE_PATH))
            os.remove(self.COMPRESS_FILE_PATH)
            self.logger.info('File {} deleted'.format(self.COMPRESS_FILE_PATH))

        self._update_task_progress(95)
        self.logger.info("Saving database timestamp {}".format(database_timestamp))

        with open(DATABASES_TS_FILE, 'w+') as f:
            f.write(str(database_timestamp))

        self.logger.info("Database timestamp saved")

        return True

    def run(self):
        ti = time.time()

        if self.create_config:
            self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(bold(self.create_config.task_name)))
            self.create_config.join()
            self.config = self.create_config.config

        self.taskman.update_progress(self.task_id, 1, self.i18n['appimage.task.db_update.checking'])

        if self.should_update(self.config):
            self.download_databases()

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        tf = time.time()
        self.logger.info("Finished. Took {0:.2f} seconds".format(tf - ti))


class SymlinksVerifier(Thread):

    def __init__(self, taskman: TaskManager, i18n: I18n, logger: logging.Logger):
        super(SymlinksVerifier, self).__init__(daemon=True)
        self.taskman = taskman
        self.i18n = i18n
        self.logger = logger
        self.task_id = 'appim_symlink_check'
        self.taskman.register_task(self.task_id, self.i18n['appimage.task.symlink_check'], get_icon_path())

    @staticmethod
    def create_symlink(app: AppImage, file_path: str, logger: logging.Logger, watcher: ProcessWatcher = None):
        logger.info("Creating a symlink for '{}'".format(app.name))
        possible_names = (app.get_clean_name(), '{}-appimage'.format(app.get_clean_name()), app.name.lower(), '{}-appimage'.format(app.name.lower()))

        if os.path.exists(SYMLINKS_DIR) and not os.path.isdir(SYMLINKS_DIR):
            logger.warning("'{}' is not a directory. It will not be possible to create a symlink for '{}'".format(SYMLINKS_DIR, app.name))
            return

        available_system_dirs = (SYMLINKS_DIR, *(l for l in ('/usr/bin', '/usr/local/bin') if os.path.isdir(l)))

        # checking if the link already exists:

        available_name = None
        for name in possible_names:
            available_name = name
            for sysdir in available_system_dirs:
                if os.path.exists('{}/{}'.format(sysdir, name)):
                    available_name = None
                    break

            if available_name:
                break

        if not available_name:
            msg = "It was not possible to create a symlink for '{}' because the names {} are already available on the system".format(app.name,
                                                                                                                                     possible_names)
            logger.warning(msg)
            if watcher:
                watcher.print('[warning] {}'.format(msg))
        else:
            try:
                Path(SYMLINKS_DIR).mkdir(parents=True, exist_ok=True)
            except Exception:
                logger.error("Could not create symlink directory '{}'".format(SYMLINKS_DIR))
                return

            symlink_path = '{}/{}'.format(SYMLINKS_DIR, available_name)

            try:
                os.symlink(src=file_path, dst=symlink_path)
                app.symlink = symlink_path

                msg = "symlink successfully created at {}".format(symlink_path)
                logger.info(msg)

                if watcher:
                    watcher.print(msg)
            except Exception:
                msg = "Could not create the symlink '{}'".format(symlink_path)
                logger.error(msg)

                if watcher:
                    watcher.print('[error] {}'.format(msg))

    def run(self):
        if os.path.exists(INSTALLATION_DIR):
            installed_files = glob.glob(f'{INSTALLATION_DIR}/*/*.json')

            if installed_files:
                self.logger.info("Checking installed AppImage files with no symlinks created")

                progress_per_file = (1/len(installed_files)) * 100
                total_progress = 0
                for json_file in installed_files:
                    with open(json_file) as f:
                        try:
                            data = json.loads(f.read())
                        except Exception:
                            self.logger.warning("Could not parse data from '{}'".format(json_file))
                            data = None

                    if data and not data.get('symlink'):
                        if not data.get('install_dir'):
                            data['install_dir'] = '/'.join(json_file.split('/')[0:-1])

                        app = AppImage(**data, i18n=self.i18n)

                        file_path = util.find_appimage_file(app.install_dir)

                        if file_path:
                            self.create_symlink(app, file_path, self.logger)
                            data['symlink'] = app.symlink

                            # caching
                            try:
                                with open(json_file, 'w+') as f:
                                    f.write(json.dumps(data))
                            except Exception:
                                self.logger.warning("Could not update cached data on '{}'".format(json_file))
                                traceback.print_exc()

                        else:
                            self.logger.warning("No AppImage file found on installation dir '{}'".format(file_path))

                    total_progress += progress_per_file
                    self.taskman.update_progress(self.task_id, total_progress, '')

                self.taskman.update_progress(self.task_id, 100, '')
                self.taskman.finish_task(self.task_id)
                return

        self.logger.info("No AppImage applications found. Aborting")
        self.taskman.update_progress(self.task_id, 100, '')
        self.taskman.finish_task(self.task_id)


class AppImageSuggestionsDownloader(Thread):

    def __init__(self, logger: logging.Logger, http_client: HttpClient, i18n: I18n, file_url: Optional[str],
                 create_config: Optional[CreateConfigFile] = None, appimage_config: Optional[dict] = None,
                 taskman: Optional[TaskManager] = None):
        super(AppImageSuggestionsDownloader, self).__init__(daemon=True)
        self.create_config = create_config
        self.logger = logger
        self.i18n = i18n
        self.http_client = http_client
        self.taskman = taskman
        self.config = appimage_config
        self.task_id = 'appim.suggestions'
        self._cached_file_path: Optional[str] = None
        self._cached_ts_file_path: Optional[str] = None

        if file_url:
            self._file_url = file_url
        else:
            self._file_url = f'https://raw.githubusercontent.com/vinifmor/bauh-files/master/appimage/suggestions.txt'

    @property
    def cached_file_path(self) -> str:
        if not self._cached_file_path:
            self._cached_file_path = f'{APPIMAGE_CACHE_DIR}/suggestions.txt'

        return self._cached_file_path

    @property
    def cached_ts_file_path(self) -> str:
        if not self._cached_ts_file_path:
            self._cached_ts_file_path = f'{APPIMAGE_CACHE_DIR}/suggestions.ts'

        return self._cached_ts_file_path

    def register_task(self):
        self.taskman.register_task(id_=self.task_id,
                                   label=self.i18n['task.download_suggestions'], icon_path=get_icon_path())

    def is_custom_local_file_mapped(self) -> bool:
        return self._file_url and self._file_url.startswith('/')

    def should_download(self, appimage_config: dict) -> bool:
        if not self._file_url:
            self.logger.error("No AppImage suggestions file URL defined")
            return False

        if self.is_custom_local_file_mapped():
            return False

        try:
            exp_hours = int(appimage_config['suggestions']['expiration'])
        except Exception:
            self.logger.error("An exception happened while trying to parse the AppImage 'suggestions.expiration'")
            traceback.print_exc()
            return True

        if exp_hours <= 0:
            self.logger.info("The AppImage suggestions cache is disabled")
            return True

        if not os.path.exists(self.cached_file_path):
            self.logger.info(f"File {self.cached_file_path} not found. It must be downloaded")
            return True

        if not os.path.exists(self.cached_ts_file_path):
            self.logger.info(f"File {self.cached_ts_file_path}  not found. The suggestions file must be downloaded.")
            return True

        with open(self.cached_ts_file_path) as f:
            timestamp_str = f.read()

        try:
            suggestions_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self.logger.error(f'Could not parse the cached AppImage suggestions timestamp: {timestamp_str}')
            traceback.print_exc()
            return True

        update = suggestions_timestamp + timedelta(hours=exp_hours) <= datetime.utcnow()
        return update

    def read(self) -> Generator[str, None, None]:
        if not self._file_url:
            self.logger.error("No AppImage suggestions file URL defined")
            yield from ()

        self.logger.info("Checking if AppImage suggestions should be downloaded")
        if self.should_download(self.config):
            suggestions_timestamp = datetime.utcnow().timestamp()
            suggestions_str = self.download()

            Thread(target=self.cache_suggestions, args=(suggestions_str, suggestions_timestamp), daemon=True).start()
        else:
            if self.is_custom_local_file_mapped():
                file_path, log_ref = self._file_url, 'local'
            else:
                file_path, log_ref = self.cached_file_path, 'cached'

            self.logger.info(f"Reading {log_ref} AppImage suggestions from {file_path}")
            with open(file_path) as f:
                suggestions_str = f.read()

        yield from self.map_suggestions(suggestions_str) if suggestions_str else ()

    def cache_suggestions(self, text: str, timestamp: float):
        self.logger.info(f"Caching AppImage suggestions to {self.cached_file_path}")

        cache_dir = os.path.dirname(self.cached_file_path)

        try:
            Path(cache_dir).mkdir(parents=True, exist_ok=True)
            cache_dir_ok = True
        except OSError:
            self.logger.error(f"Could not create the caching directory {cache_dir}")
            traceback.print_exc()
            cache_dir_ok = False

        if cache_dir_ok:
            try:
                with open(self.cached_file_path, 'w+') as f:
                    f.write(text)
            except Exception:
                self.logger.error(f"An exception happened while writing AppImage suggestions to {self.cached_file_path}")
                traceback.print_exc()

            try:
                with open(self.cached_ts_file_path, 'w+') as f:
                    f.write(str(timestamp))
            except Exception:
                self.logger.error(f"An exception happened while writing the cached AppImage suggestions timestamp "
                                  f"to {self.cached_ts_file_path}")
                traceback.print_exc()

    def download(self) -> Optional[str]:
        if not self._file_url:
            self.logger.error("No AppImage suggestions file URL defined")
            return

        if self.is_custom_local_file_mapped():
            self.logger.warning("Local AppImage suggestions file mapped. Nothing will be downloaded")
            return

        self.logger.info(f"Downloading AppImage suggestions from {self._file_url}")

        try:
            res = self.http_client.get(self._file_url)
        except requests.exceptions.ConnectionError:
            self.logger.warning(f"Could not download suggestion from {self._file_url}")
            return

        if not res:
            self.logger.warning(f"Could not download suggestion from {self._file_url}")
            return

        if not res.text:
            self.logger.warning(f"No AppImage suggestion found in {self._file_url}")
            return

        return res.text

    def map_suggestions(self, text: str) -> Generator[str, None, None]:
        return (line for line in text.split('\n') if line)

    def run(self):
        ti = time.time()

        if not self.is_custom_local_file_mapped():
            if self.create_config:
                wait_msg = self.i18n['task.waiting_task'].format(bold(self.create_config.task_name))
                self.taskman.update_progress(self.task_id, 0, wait_msg)
                self.create_config.join()
                self.config = self.create_config.config

            ti = time.time()
            self.taskman.update_progress(self.task_id, 1, None)

            self.logger.info("Checking if AppImage suggestions should be downloaded")
            should_download = self.should_download(self.config)
            self.taskman.update_progress(self.task_id, 30, None)

            try:
                if should_download:
                    suggestions_timestamp = datetime.utcnow().timestamp()
                    suggestions_str = self.download()
                    self.taskman.update_progress(self.task_id, 70, None)

                    if suggestions_str:
                        self.cache_suggestions(suggestions_str, suggestions_timestamp)
                else:
                    self.logger.info("Cached AppImage suggestions are up-to-date")
            except Exception:
                self.logger.error("An unexpected exception happened while downloading AppImage suggestions")
                traceback.print_exc()

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        tf = time.time()
        self.logger.info(f"Took {tf - ti:.9f} seconds to download suggestions")



from getpass import getuser
from pathlib import Path

from bauh import __app_name__
from bauh.api import user


def get_temp_dir(username: str) -> str:
    return f'/tmp/{__app_name__}@{username}'


CACHE_DIR = f'/var/cache/{__app_name__}' if user.is_root() else f'{Path.home()}/.cache/{__app_name__}'
CONFIG_DIR = f'/etc/{__app_name__}' if user.is_root() else f'{Path.home()}/.config/{__app_name__}'
USER_THEMES_DIR = f'/usr/share/{__app_name__}/themes' if user.is_root() else f'{Path.home()}/.local/share/{__app_name__}/themes'
DESKTOP_ENTRIES_DIR = '/usr/share/applications' if user.is_root() else f'{Path.home()}/.local/share/applications'
TEMP_DIR = get_temp_dir(getuser())
LOGS_DIR = f'{TEMP_DIR}/logs'
AUTOSTART_DIR = f'/etc/xdg/autostart' if user.is_root() else f'{Path.home()}/.config/autostart'
BINARIES_DIR = f'/usr/local/bin' if user.is_root() else f'{Path.home()}/.local/bin'
SHARED_FILES_DIR = f'/usr/local/share/{__app_name__}' if user.is_root() else f'{Path.home()}/.local/share/{__app_name__}'



import os
from typing import Optional, Dict
from bauh import __app_name__


def read_suggestions_mapping() -> Optional[Dict[str, str]]:
    file_path = f'/etc/{__app_name__}/suggestions.conf'

    if os.path.isfile(file_path):
        try:
            with open(file_path) as f:
                file_content = f.read()
        except FileNotFoundError:
            return

        if not file_content:
            return

        mapping = {}
        for line in file_content.split('\n'):
            line_strip = line.strip()

            if not line_strip.startswith('#'):
                gem_file = line_strip.split('=')

                if len(gem_file) == 2:
                    gem_name, file_url = gem_file[0].strip(), gem_file[1].strip()

                    if gem_name and file_url:
                        mapping[gem_name] = file_url

        return mapping if mapping else None



import os
import re
import subprocess
import sys
import time
from io import StringIO
from subprocess import PIPE
from typing import List, Tuple, Set, Dict, Optional, Iterable, Union, IO, Any

# default environment variables for subprocesses.
from bauh.api.abstract.handler import ProcessWatcher

PY_VERSION = "{}.{}".format(sys.version_info.major, sys.version_info.minor)
GLOBAL_PY_LIBS = '/usr/lib/python{}'.format(PY_VERSION)

PATH = os.getenv('PATH')
DEFAULT_LANG = ''

GLOBAL_INTERPRETER_PATH = ':'.join(PATH.split(':')[1:])

if GLOBAL_PY_LIBS not in PATH:
    PATH = '{}:{}'.format(GLOBAL_PY_LIBS, PATH)

USE_GLOBAL_INTERPRETER = bool(os.getenv('VIRTUAL_ENV'))

RE_SUDO_OUTPUT = re.compile(r'[sudo]\s*[\w\s]+:\s*')


def gen_env(global_interpreter: bool = USE_GLOBAL_INTERPRETER, lang: Optional[str] = DEFAULT_LANG,
            extra_paths: Optional[Set[str]] = None) -> dict:

    custom_env = dict(os.environ)

    if lang is not None:
        custom_env['LANG'] = lang

    if global_interpreter:  # to avoid subprocess calls to the virtualenv python interpreter instead of the global one.
        custom_env['PATH'] = GLOBAL_INTERPRETER_PATH

    else:
        custom_env['PATH'] = PATH

    if extra_paths:
        custom_env['PATH'] = ':'.join(extra_paths) + ':' + custom_env['PATH']

    return custom_env


class SystemProcess:

    """
    Represents a system process being executed.
    """

    def __init__(self, subproc: subprocess.Popen, success_phrases: List[str] = None, wrong_error_phrase: str = '[sudo] password for',
                 check_error_output: bool = True, skip_stdout: bool = False, output_delay: float = None):
        self.subproc = subproc
        self.success_phrases = success_phrases
        self.wrong_error_phrase = wrong_error_phrase
        self.check_error_output = check_error_output
        self.skip_stdout = skip_stdout
        self.output_delay = output_delay

    def wait(self):
        self.subproc.wait()


class SimpleProcess:

    def __init__(self, cmd: Iterable[str], cwd: str = '.', expected_code: int = 0,
                 global_interpreter: bool = USE_GLOBAL_INTERPRETER, lang: Optional[str] = DEFAULT_LANG, root_password: Optional[str] = None,
                 extra_paths: Set[str] = None, error_phrases: Set[str] = None, wrong_error_phrases: Set[str] = None,
                 shell: bool = False, success_phrases: Set[str] = None, extra_env: Optional[Dict[str, str]] = None,
                 custom_user: Optional[str] = None, preserve_env: Optional[Set] = None):
        pwdin, final_cmd = None, []

        self.shell = shell

        if custom_user:
            final_cmd.extend(['runuser', '-u', custom_user, '--'])
        elif isinstance(root_password, str):
            final_cmd.extend(['sudo', '-S'])

            if preserve_env:
                for var in preserve_env:
                    final_cmd.append(f'--preserve-env={var}')

            pwdin = self._new(['echo', root_password], cwd, global_interpreter, lang).stdout

        final_cmd.extend(cmd)

        self.instance = self._new(final_cmd, cwd, global_interpreter, lang=lang, stdin=pwdin,
                                  extra_paths=extra_paths, extra_env=extra_env)
        self.expected_code = expected_code
        self.error_phrases = error_phrases
        self.wrong_error_phrases = wrong_error_phrases
        self.success_phrases = success_phrases

    def _new(self, cmd: List[str], cwd: str, global_interpreter: bool, lang: Optional[str], stdin = None,
             extra_paths: Set[str] = None, extra_env: Optional[Dict[str, str]] = None) -> subprocess.Popen:

        env = gen_env(global_interpreter=global_interpreter, lang=lang, extra_paths=extra_paths)

        if extra_env:
            for var, val in extra_env.items():
                env[var] = val

        args = {
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": stdin if stdin else subprocess.DEVNULL,
            "bufsize": -1,
            "cwd": cwd,
            "env": env,
            "shell": self.shell
        }

        return subprocess.Popen(args=[' '.join(cmd)] if self.shell else cmd, **args)


class ProcessHandler:

    """
    It handles a process execution and notifies a specified watcher.
    """

    def __init__(self, watcher: ProcessWatcher = None):
        self.watcher = watcher

    def _notify_watcher(self, msg: str, as_substatus: bool = False):
        if self.watcher:
            self.watcher.print(msg)

            if as_substatus:
                self.watcher.change_substatus(msg)

    def handle(self, process: SystemProcess, error_output: StringIO = None, output_handler=None) -> bool:
        self._notify_watcher(' '.join(process.subproc.args) + '\n')

        already_succeeded = False

        if not process.skip_stdout:
            for output in process.subproc.stdout:

                try:
                    line = output.decode().strip()
                except UnicodeDecodeError:
                    line = None

                if line:
                    self._notify_watcher(line)

                    if output_handler:
                        output_handler(line)

                    if process.success_phrases and [p in line for p in process.success_phrases]:
                        already_succeeded = True

                if not already_succeeded and process.output_delay:
                    time.sleep(process.output_delay)

            if already_succeeded:
                return True

        for output in process.subproc.stderr:
            if output:
                try:
                    line = output.decode().strip()
                except UnicodeDecodeError:
                    line = None

                if line:
                    self._notify_watcher(line)

                    if output_handler:
                        output_handler(line)

                    if error_output is not None:
                        error_output.write(line)

                    if process.check_error_output:
                        if process.wrong_error_phrase and process.wrong_error_phrase in line:
                            continue
                        else:
                            return False
                    elif process.skip_stdout and process.success_phrases and [p in line for p in process.success_phrases]:
                        already_succeeded = True

                if not already_succeeded and process.output_delay:
                    time.sleep(process.output_delay)

        if already_succeeded:
            return True

        return process.subproc.returncode is None or process.subproc.returncode == 0

    def handle_simple(self, proc: SimpleProcess, output_handler=None, notify_watcher: bool = True,
                      output_as_substatus: bool = False) -> Tuple[bool, str]:
        if notify_watcher:
            self._notify_watcher((proc.instance.args if isinstance(proc.instance.args, str) else ' '.join(proc.instance.args)) + '\n')

        output = StringIO()
        for o in proc.instance.stdout:
            if o:
                try:
                    line = o.decode()
                except UnicodeDecodeError:
                    continue

                if line.startswith('[sudo]'):
                    line = RE_SUDO_OUTPUT.split(line)[1]

                output.write(line)

                line = line.strip()

                if line:
                    if output_handler:
                        output_handler(line)

                    if notify_watcher:
                        self._notify_watcher(line, as_substatus=output_as_substatus)

        proc.instance.wait()
        output.seek(0)

        success = proc.instance.returncode == proc.expected_code
        string_output = output.read()

        if proc.success_phrases:
            for phrase in proc.success_phrases:
                if phrase in string_output:
                    success = True
                    break

        if not success and proc.wrong_error_phrases:
            for phrase in proc.wrong_error_phrases:
                if phrase in string_output:
                    success = True
                    break

        if success and proc.error_phrases:
            for phrase in proc.error_phrases:
                if phrase in string_output:
                    success = False
                    break

        return success, string_output


def run_cmd(cmd: str, expected_code: int = 0, ignore_return_code: bool = False, print_error: bool = True,
            cwd: str = '.', global_interpreter: bool = USE_GLOBAL_INTERPRETER, extra_paths: Set[str] = None,
            custom_user: Optional[str] = None, lang: Optional[str] = DEFAULT_LANG) -> Optional[str]:
    """
    runs a given command and returns its default output
    :return:
    """
    args = {
        "shell": True,
        "stdout": PIPE,
        "env": gen_env(global_interpreter=global_interpreter, lang=lang, extra_paths=extra_paths),
        'cwd': cwd
    }

    if not print_error:
        args["stderr"] = subprocess.DEVNULL

    final_cmd = f"runuser -u {custom_user} -- {cmd}" if custom_user else cmd
    res = subprocess.run(final_cmd, **args)

    if ignore_return_code or res.returncode == expected_code:
        try:
            return res.stdout.decode()
        except UnicodeDecodeError:
            pass


def new_subprocess(cmd: Iterable[str], cwd: str = '.', shell: bool = False, stdin: Optional[Union[None, int, IO[Any]]] = None,
                   global_interpreter: bool = USE_GLOBAL_INTERPRETER, lang: Optional[str] = DEFAULT_LANG,
                   extra_paths: Set[str] = None, custom_user: Optional[str] = None) -> subprocess.Popen:
    args = {
        "stdout": PIPE,
        "stderr": PIPE,
        "cwd": cwd,
        "shell": shell,
        "env": gen_env(global_interpreter, lang, extra_paths),
        "stdin": stdin if stdin else subprocess.DEVNULL
    }

    final_cmd = ['runuser', '-u', custom_user, '--', *cmd] if custom_user else cmd
    return subprocess.Popen(final_cmd, **args)


def new_root_subprocess(cmd: Iterable[str], root_password: Optional[str], cwd: str = '.',
                        global_interpreter: bool = USE_GLOBAL_INTERPRETER, lang: str = DEFAULT_LANG,
                        extra_paths: Set[str] = None, shell: bool = False) -> subprocess.Popen:
    pwdin, final_cmd = subprocess.DEVNULL, []

    if isinstance(root_password, str):
        final_cmd.extend(['sudo', '-S'])
        pwdin = new_subprocess(['echo', root_password], global_interpreter=global_interpreter, lang=lang).stdout

    final_cmd.extend(cmd)

    if shell:
        final_cmd = ' '.join(final_cmd)

    return subprocess.Popen(final_cmd, stdin=pwdin, stdout=PIPE, stderr=PIPE, cwd=cwd,
                            env=gen_env(global_interpreter, lang, extra_paths), shell=shell)


def notify_user(msg: str, app_name: str, icon_path: str):
    os.system("notify-send -a {} {} '{}'".format(app_name, "-i {}".format(icon_path) if icon_path else '', msg))


def get_dir_size(start_path='.'):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)

            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


def run(cmd: List[str], success_code: int = 0, custom_user: Optional[str] = None) -> Tuple[bool, str]:
    final_cmd = ['runuser', '-u', custom_user, '--', *cmd] if custom_user else cmd
    p = subprocess.run(final_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.DEVNULL)

    try:
        output = p.stdout.decode()
    except UnicodeDecodeError:
        output = ''

    return p.returncode == success_code, output


def check_active_services(*names: str) -> Dict[str, bool]:
    output = run_cmd('systemctl is-active {}'.format(' '.join(names)), print_error=False)

    if not output:
        return {n: False for n in names}
    else:
        status = output.split('\n')
        return {s: status[i].strip().lower() == 'active' for i, s in enumerate(names) if s}


def check_enabled_services(*names: str) -> Dict[str, bool]:
    output = run_cmd('systemctl is-enabled {}'.format(' '.join(names)), print_error=False)

    if not output:
        return {n: False for n in names}
    else:
        status = output.split('\n')
        return {s: status[i].strip().lower() == 'enabled' for i, s in enumerate(names) if s}


def execute(cmd: str, shell: bool = False, cwd: Optional[str] = None, output: bool = True, custom_env: Optional[dict] = None,
            stdin: bool = True, custom_user: Optional[str] = None) -> Tuple[int, Optional[str]]:

    final_cmd = f"runuser -u {custom_user} -- {cmd}" if custom_user else cmd

    params = {
        'args': final_cmd.split(' ') if not shell else [final_cmd],
        'stdout': subprocess.PIPE if output else subprocess.DEVNULL,
        'stderr': subprocess.STDOUT if output else subprocess.DEVNULL,
        'shell': shell
    }

    if not stdin:
        params['stdin'] = subprocess.DEVNULL

    if cwd is not None:
        params['cwd'] = cwd

    if custom_env is not None:
        params['env'] = custom_env

    p = subprocess.run(**params)

    output = None
    if p.stdout:
        try:
            output = p.stdout.decode()
        except UnicodeDecodeError:
            output = None

    return p.returncode, output



import logging
import os
import time
import traceback
from datetime import datetime, timedelta
from pathlib import Path
from threading import Thread
from typing import Dict, List, Optional

import requests

from bauh.api.abstract.controller import SoftwareManager
from bauh.api.http import HttpClient
from bauh.commons.internet import InternetChecker
from bauh.commons.util import map_timestamp_file


class CategoriesDownloader(Thread):

    def __init__(self, id_: str, http_client: HttpClient, logger: logging.Logger, manager: SoftwareManager,
                 url_categories_file: str, categories_path: str, internet_checker: InternetChecker,
                 expiration: Optional[int] = None, internet_connection: Optional[bool] = True, before=None, after=None):
        """
        :param id_:
        :param http_client:
        :param logger:
        :param manager:
        :param url_categories_file:
        :param categories_path:
        :param expiration: cached file expiration in hours
        :param internet_checker
        :param before:
        :param after:
        """
        super(CategoriesDownloader, self).__init__(daemon=True)
        self.id_ = id_
        self.http_client = http_client
        self.logger = logger
        self.manager = manager
        self.url_categories_file = url_categories_file
        self.categories_path = categories_path
        self.before = before
        self.after = after
        self.expiration = expiration
        self.internet_connection = internet_connection
        self.internet_checker = internet_checker

    def _msg(self, msg: str):
        return '{} [{}]: {}'.format(self.__class__.__name__, self.id_, msg)

    def _read_categories_from_disk(self) -> Dict[str, List[str]]:
        if os.path.exists(self.categories_path):
            self.logger.info(self._msg("Reading cached categories file {}".format(self.categories_path)))

            with open(self.categories_path) as f:
                categories = f.read()

            return self._map_categories(categories)
        else:
            self.logger.warning("No cached categories file {} found".format(self.categories_path))
            return {}

    def _map_categories(self, categories: str) -> Dict[str, List[str]]:
        categories_map = {}
        for l in categories.split('\n'):
            if l:
                data = l.split('=')
                categories_map[data[0]] = [c.strip() for c in data[1].split(',') if c]

        return categories_map

    def _cache_categories_to_disk(self, categories_str: str, timestamp: float):
        self.logger.info(self._msg('Caching downloaded categories to disk'))

        try:
            Path(os.path.dirname(self.categories_path)).mkdir(parents=True, exist_ok=True)

            with open(self.categories_path, 'w+') as f:
                f.write(categories_str)

            self.logger.info(self._msg("Categories cached to file '{}'".format(self.categories_path)))

            categories_ts_path = map_timestamp_file(self.categories_path)
            with open(categories_ts_path, 'w+') as f:
                f.write(str(timestamp))

            self.logger.info(self._msg("Categories timestamp ({}) cached to file '{}'".format(timestamp, categories_ts_path)))
        except Exception:
            self.logger.error(self._msg("Could not cache categories to the disk as '{}'".format(self.categories_path)))
            traceback.print_exc()

    def download_categories(self) -> Dict[str, List[str]]:
        self.logger.info(self._msg('Downloading category definitions from {}'.format(self.url_categories_file)))

        try:
            timestamp = datetime.utcnow().timestamp()
            res = self.http_client.get(self.url_categories_file)
        except requests.exceptions.ConnectionError:
            self.logger.error(self._msg('[{}] Could not download categories. The internet connection seems to be off.'.format(self.id_)))
            return {}

        if not res:
            self.logger.info(self._msg('Could not download {}'.format(self.url_categories_file)))
            return {}

        try:
            categories = self._map_categories(res.text)
            self.logger.info(self._msg('Loaded categories for {} applications'.format(len(categories))))
        except Exception:
            self.logger.error(self._msg("Could not parse categories definitions"))
            traceback.print_exc()
            return {}

        if categories:
            self._cache_categories_to_disk(categories_str=res.text, timestamp=timestamp)

        return categories

    def should_download(self) -> bool:
        if self.internet_connection is False or (self.internet_connection is None and not self.internet_checker.is_available()):
            self.logger.warning(self._msg("No internet connection. The categories file '{}' cannot be updated.".format(self.categories_path)))
            return False

        if self.expiration is None or self.expiration <= 0:
            self.logger.warning(self._msg("No expiration set for the categories file '{}'. It should be downloaded".format(self.categories_path)))
            return True

        if not os.path.exists(self.categories_path):
            self.logger.warning(self._msg("Categories file '{}' does not exist. It should be downloaded.".format(self.categories_path)))
            return True

        categories_ts_path = map_timestamp_file(self.categories_path)

        if not os.path.exists(categories_ts_path):
            self.logger.warning(self._msg("Categories timestamp file '{}' does not exist. The categories file should be re-downloaded.".format(categories_ts_path)))
            return True

        with open(categories_ts_path) as f:
            timestamp_str = f.read()

        try:
            categories_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self.logger.error(self._msg("An exception occurred when trying to parse the categories file timestamp from '{}'. The categories file should be re-downloaded.".format(categories_ts_path)))
            traceback.print_exc()
            return True

        should_download = (categories_timestamp + timedelta(hours=self.expiration) <= datetime.utcnow())

        if should_download:
            self.logger.info(self._msg("Cached categories file '{}' has expired. A new one should be downloaded.".format(self.categories_path)))
            return True
        else:
            self.logger.info(self._msg("Cached categories file '{}' is up to date. No need to re-download it.".format(self.categories_path)))
            return False

    def run(self):
        ti = time.time()
        if self.before:
            self.before()

        should_download = self.should_download()

        if not should_download:
            cached = self._read_categories_from_disk()
            self.manager.categories = cached
        else:
            self.download_categories()

        if self.after:
            self.after()

        tf = time.time()
        self.logger.info(self._msg('Finished. Took {0:.2f} seconds'.format(tf - ti)))
